# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2006-2008, 2011 Lukáš Lalinský
# Copyright (C) 2009, 2011, 2019-2021 Philipp Wolfer
# Copyright (C) 2011-2013 Michael Wiencek
# Copyright (C) 2018, 2020-2021 Laurent Monin
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from picard.config import (
    FloatOption,
    get_config,
)

from picard.ui.options import (
    OptionsPage,
    register_options_page,
)
from picard.ui.ui_options_matching import Ui_MatchingOptionsPage


class MatchingOptionsPage(OptionsPage):

    NAME = 'matching'
    TITLE = N_("Matching")
    PARENT = 'advanced'
    SORT_ORDER = 30
    ACTIVE = True
    HELP_URL = "/config/options_matching.html"

    options = [
        FloatOption('setting', 'file_lookup_threshold', 0.7),
        FloatOption('setting', 'cluster_lookup_threshold', 0.7),
        FloatOption('setting', 'track_matching_threshold', 0.4),
    ]

    _release_type_sliders = {}

    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_MatchingOptionsPage()
        self.ui.setupUi(self)

    def load(self):
        config = get_config()
        self.ui.file_lookup_threshold.setValue(int(config.setting['file_lookup_threshold'] * 100))
        self.ui.cluster_lookup_threshold.setValue(int(config.setting['cluster_lookup_threshold'] * 100))
        self.ui.track_matching_threshold.setValue(int(config.setting['track_matching_threshold'] * 100))

    def save(self):
        config = get_config()
        config.setting['file_lookup_threshold'] = float(self.ui.file_lookup_threshold.value()) / 100.0
        config.setting['cluster_lookup_threshold'] = float(self.ui.cluster_lookup_threshold.value()) / 100.0
        config.setting['track_matching_threshold'] = float(self.ui.track_matching_threshold.value()) / 100.0


register_options_page(MatchingOptionsPage)



# -*- coding: utf-8 -*-

# Automatically generated - don't edit.
# Use `python setup.py build_ui` to update it.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_WinCompatDialog(object):
    def setupUi(self, WinCompatDialog):
        WinCompatDialog.setObjectName("WinCompatDialog")
        WinCompatDialog.setWindowModality(QtCore.Qt.WindowModal)
        WinCompatDialog.resize(285, 295)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(WinCompatDialog.sizePolicy().hasHeightForWidth())
        WinCompatDialog.setSizePolicy(sizePolicy)
        self.verticalLayout = QtWidgets.QVBoxLayout(WinCompatDialog)
        self.verticalLayout.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.label_header_character = QtWidgets.QLabel(WinCompatDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_header_character.setFont(font)
        self.label_header_character.setObjectName("label_header_character")
        self.gridLayout.addWidget(self.label_header_character, 0, 0, 1, 1)
        self.label_header_replace = QtWidgets.QLabel(WinCompatDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_header_replace.sizePolicy().hasHeightForWidth())
        self.label_header_replace.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_header_replace.setFont(font)
        self.label_header_replace.setObjectName("label_header_replace")
        self.gridLayout.addWidget(self.label_header_replace, 0, 2, 1, 1)
        self.label_lt = QtWidgets.QLabel(WinCompatDialog)
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.label_lt.setFont(font)
        self.label_lt.setText("<")
        self.label_lt.setObjectName("label_lt")
        self.gridLayout.addWidget(self.label_lt, 3, 0, 1, 1)
        self.label_colon = QtWidgets.QLabel(WinCompatDialog)
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.label_colon.setFont(font)
        self.label_colon.setText(":")
        self.label_colon.setObjectName("label_colon")
        self.gridLayout.addWidget(self.label_colon, 2, 0, 1, 1)
        self.label_gt = QtWidgets.QLabel(WinCompatDialog)
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.label_gt.setFont(font)
        self.label_gt.setText(">")
        self.label_gt.setObjectName("label_gt")
        self.gridLayout.addWidget(self.label_gt, 4, 0, 1, 1)
        self.replace_questionmark = QtWidgets.QLineEdit(WinCompatDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.replace_questionmark.sizePolicy().hasHeightForWidth())
        self.replace_questionmark.setSizePolicy(sizePolicy)
        self.replace_questionmark.setMaximumSize(QtCore.QSize(20, 16777215))
        self.replace_questionmark.setText("_")
        self.replace_questionmark.setMaxLength(1)
        self.replace_questionmark.setObjectName("replace_questionmark")
        self.gridLayout.addWidget(self.replace_questionmark, 5, 2, 1, 1)
        self.label_questionmark = QtWidgets.QLabel(WinCompatDialog)
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.label_questionmark.setFont(font)
        self.label_questionmark.setText("?")
        self.label_questionmark.setObjectName("label_questionmark")
        self.gridLayout.addWidget(self.label_questionmark, 5, 0, 1, 1)
        self.label_pipe = QtWidgets.QLabel(WinCompatDialog)
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.label_pipe.setFont(font)
        self.label_pipe.setText("|")
        self.label_pipe.setObjectName("label_pipe")
        self.gridLayout.addWidget(self.label_pipe, 6, 0, 1, 1)
        self.replace_asterisk = QtWidgets.QLineEdit(WinCompatDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.replace_asterisk.sizePolicy().hasHeightForWidth())
        self.replace_asterisk.setSizePolicy(sizePolicy)
        self.replace_asterisk.setMaximumSize(QtCore.QSize(20, 16777215))
        self.replace_asterisk.setText("_")
        self.replace_asterisk.setMaxLength(1)
        self.replace_asterisk.setObjectName("replace_asterisk")
        self.gridLayout.addWidget(self.replace_asterisk, 1, 2, 1, 1)
        self.replace_gt = QtWidgets.QLineEdit(WinCompatDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.replace_gt.sizePolicy().hasHeightForWidth())
        self.replace_gt.setSizePolicy(sizePolicy)
        self.replace_gt.setMaximumSize(QtCore.QSize(20, 16777215))
        self.replace_gt.setText("_")
        self.replace_gt.setMaxLength(1)
        self.replace_gt.setObjectName("replace_gt")
        self.gridLayout.addWidget(self.replace_gt, 4, 2, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 0, 3, 1, 1)
        self.replace_lt = QtWidgets.QLineEdit(WinCompatDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.replace_lt.sizePolicy().hasHeightForWidth())
        self.replace_lt.setSizePolicy(sizePolicy)
        self.replace_lt.setMaximumSize(QtCore.QSize(20, 16777215))
        self.replace_lt.setText("_")
        self.replace_lt.setMaxLength(1)
        self.replace_lt.setObjectName("replace_lt")
        self.gridLayout.addWidget(self.replace_lt, 3, 2, 1, 1)
        self.label_asterisk = QtWidgets.QLabel(WinCompatDialog)
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.label_asterisk.setFont(font)
        self.label_asterisk.setText("*")
        self.label_asterisk.setObjectName("label_asterisk")
        self.gridLayout.addWidget(self.label_asterisk, 1, 0, 1, 1)
        self.replace_pipe = QtWidgets.QLineEdit(WinCompatDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.replace_pipe.sizePolicy().hasHeightForWidth())
        self.replace_pipe.setSizePolicy(sizePolicy)
        self.replace_pipe.setMaximumSize(QtCore.QSize(20, 16777215))
        self.replace_pipe.setText("_")
        self.replace_pipe.setMaxLength(1)
        self.replace_pipe.setObjectName("replace_pipe")
        self.gridLayout.addWidget(self.replace_pipe, 6, 2, 1, 1)
        self.replace_colon = QtWidgets.QLineEdit(WinCompatDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.replace_colon.sizePolicy().hasHeightForWidth())
        self.replace_colon.setSizePolicy(sizePolicy)
        self.replace_colon.setMaximumSize(QtCore.QSize(20, 16777215))
        self.replace_colon.setText("_")
        self.replace_colon.setMaxLength(1)
        self.replace_colon.setObjectName("replace_colon")
        self.gridLayout.addWidget(self.replace_colon, 2, 2, 1, 1)
        self.label_quotationmark = QtWidgets.QLabel(WinCompatDialog)
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.label_quotationmark.setFont(font)
        self.label_quotationmark.setText("\"")
        self.label_quotationmark.setObjectName("label_quotationmark")
        self.gridLayout.addWidget(self.label_quotationmark, 7, 0, 1, 1)
        self.replace_quotationmark = QtWidgets.QLineEdit(WinCompatDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.replace_quotationmark.sizePolicy().hasHeightForWidth())
        self.replace_quotationmark.setSizePolicy(sizePolicy)
        self.replace_quotationmark.setMaximumSize(QtCore.QSize(20, 16777215))
        self.replace_quotationmark.setText("_")
        self.replace_quotationmark.setObjectName("replace_quotationmark")
        self.gridLayout.addWidget(self.replace_quotationmark, 7, 2, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem1)
        self.buttonbox = QtWidgets.QDialogButtonBox(WinCompatDialog)
        self.buttonbox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonbox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Save)
        self.buttonbox.setObjectName("buttonbox")
        self.verticalLayout.addWidget(self.buttonbox)

        self.retranslateUi(WinCompatDialog)
        QtCore.QMetaObject.connectSlotsByName(WinCompatDialog)
        WinCompatDialog.setTabOrder(self.replace_asterisk, self.replace_colon)
        WinCompatDialog.setTabOrder(self.replace_colon, self.replace_lt)
        WinCompatDialog.setTabOrder(self.replace_lt, self.replace_gt)
        WinCompatDialog.setTabOrder(self.replace_gt, self.replace_questionmark)
        WinCompatDialog.setTabOrder(self.replace_questionmark, self.replace_pipe)
        WinCompatDialog.setTabOrder(self.replace_pipe, self.replace_quotationmark)

    def retranslateUi(self, WinCompatDialog):
        _translate = QtCore.QCoreApplication.translate
        WinCompatDialog.setWindowTitle(_("Windows compatibility"))
        self.label_header_character.setText(_("Character"))
        self.label_header_replace.setText(_("Replacement"))



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2013, 2018, 2020-2022 Laurent Monin
# Copyright (C) 2016-2017 Sambhav Kothari
# Copyright (C) 2019, 2021-2022 Philipp Wolfer
# Copyright (C) 2021 Gabriel Ferreira
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


import time

from PyQt5 import (
    QtCore,
    QtGui,
    QtWidgets,
)

from picard.util import icontheme
from picard.util.time import get_timestamp

from picard.ui.ui_infostatus import Ui_InfoStatus


class InfoStatus(QtWidgets.QWidget, Ui_InfoStatus):

    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self, parent)
        Ui_InfoStatus.__init__(self)
        self.setupUi(self)

        self._size = QtCore.QSize(16, 16)
        self._create_icons()
        self._init_labels()

        self.reset_counters()

    def _init_labels(self):
        size = self._size
        self.label1.setPixmap(self.icon_eta.pixmap(size))
        self.label1.hide()
        self.label2.setPixmap(self.icon_file.pixmap(size))
        self.label3.setPixmap(self.icon_cd.pixmap(size))
        self.label4.setPixmap(self.icon_file_pending.pixmap(size))
        self.label5.setPixmap(self.icon_download.pixmap(size, QtGui.QIcon.Mode.Disabled))
        self._init_tooltips()

    def _create_icons(self):
        self.icon_eta = QtGui.QIcon(":/images/22x22/hourglass.png")
        self.icon_cd = icontheme.lookup('media-optical')
        self.icon_file = QtGui.QIcon(":/images/file.png")
        self.icon_file_pending = QtGui.QIcon(":/images/file-pending.png")
        self.icon_download = QtGui.QIcon(":/images/16x16/action-go-down-16.png")

    def _init_tooltips(self):
        t1 = _("Estimated Time")
        t2 = _("Files")
        t3 = _("Albums")
        t4 = _("Pending files")
        t5 = _("Pending requests")
        self.val1.setToolTip(t1)
        self.label1.setToolTip(t1)
        self.val2.setToolTip(t2)
        self.label2.setToolTip(t2)
        self.val3.setToolTip(t3)
        self.label3.setToolTip(t3)
        self.val4.setToolTip(t4)
        self.label4.setToolTip(t4)
        self.val5.setToolTip(t5)
        self.label5.setToolTip(t5)

    def update(self, files=0, albums=0, pending_files=0, pending_requests=0, progress=0):
        self.set_files(files)
        self.set_albums(albums)
        self.set_pending_files(pending_files)
        self.set_pending_requests(pending_requests)

        # estimate eta
        total_pending = pending_files + pending_requests
        last_pending = self._last_pending_files + self._last_pending_requests

        # Reset the counters if we had no pending progress before and receive new pending items.
        # This resets the starting timestamp and starts a new round of measurement.
        if total_pending > 0 and last_pending == 0:
            self.reset_counters()

        previous_done_files = max(0, self._max_pending_files - self._last_pending_files)
        previous_done_requests = max(0, self._max_pending_requests - self._last_pending_requests)
        self._max_pending_files = max(self._max_pending_files, previous_done_files + pending_files)
        self._max_pending_requests = max(self._max_pending_requests, previous_done_requests + pending_requests)
        self._last_pending_files = pending_files
        self._last_pending_requests = pending_requests

        if total_pending == 0 or (self._max_pending_files + self._max_pending_requests <= 1):
            self.reset_counters()
            self.hide_eta()
            return

        if total_pending != last_pending:
            current_time = time.time()

            # time since we started processing this batch
            diff_time = max(0.1, current_time - self._prev_time)  # denominator can't be 0
            previous_done_files = max(1, previous_done_files)  # denominator can't be 0

            # we estimate based on the time per file * number of pending files + 1 second per additional request
            file_eta_seconds = (diff_time / previous_done_files) * pending_files + pending_requests

            # we assume additional network requests based on the ratio of requests/files * pending files
            # to estimate an upper bound (e.g. fetch cover, lookup, scan)
            network_eta_seconds = pending_requests + (previous_done_requests / previous_done_files) * pending_files

            # general eta (biased towards whatever takes longer)
            eta_seconds = max(network_eta_seconds, file_eta_seconds)

            # estimate progress
            self._last_progress = diff_time / (diff_time + eta_seconds)
            self.set_eta(eta_seconds)

    def reset_counters(self):
        self._last_progress = 0
        self._max_pending_requests = 0
        self._last_pending_requests = 0
        self._max_pending_files = 0
        self._last_pending_files = 0
        self._prev_time = time.time()

    def get_progress(self):
        return self._last_progress

    def set_eta(self, eta_seconds):
        if eta_seconds > 0:
            self.val1.setText(get_timestamp(eta_seconds))
            self.val1.show()
            self.label1.show()
        else:
            self.hide_eta()

    def hide_eta(self):
        self.val1.hide()
        self.label1.hide()

    def set_files(self, num):
        self.val2.setText(str(num))

    def set_albums(self, num):
        self.val3.setText(str(num))

    def set_pending_files(self, num):
        self.val4.setText(str(num))

    def set_pending_requests(self, num):
        if num <= 0:
            enabled = QtGui.QIcon.Mode.Disabled
        else:
            enabled = QtGui.QIcon.Mode.Normal
        self.label5.setPixmap(self.icon_download.pixmap(self._size, enabled))
        self.val5.setText(str(num))



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2006-2008, 2012 Lukáš Lalinský
# Copyright (C) 2008 Hendrik van Antwerpen
# Copyright (C) 2008-2010, 2014-2015, 2018-2023 Philipp Wolfer
# Copyright (C) 2012-2013 Michael Wiencek
# Copyright (C) 2012-2014 Wieland Hoffmann
# Copyright (C) 2013 Calvin Walton
# Copyright (C) 2013-2014, 2017-2023 Laurent Monin
# Copyright (C) 2016-2018 Sambhav Kothari
# Copyright (C) 2017 Ville Skyttä
# Copyright (C) 2022 Marcin Szalowicz
# Copyright (C) 2023 certuna
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


import base64
import re

import mutagen.flac
import mutagen.ogg
import mutagen.oggflac
import mutagen.oggopus
import mutagen.oggspeex
import mutagen.oggtheora
import mutagen.oggvorbis

from picard import log
from picard.config import get_config
from picard.coverart.image import (
    CoverArtImageError,
    TagCoverArtImage,
)
from picard.coverart.utils import types_from_id3
from picard.file import File
from picard.formats.util import guess_format
from picard.metadata import Metadata
from picard.util import (
    encode_filename,
    sanitize_date,
)


FLAC_MAX_BLOCK_SIZE = 2 ** 24 - 1  # FLAC block size is limited to a 24 bit integer
INVALID_CHARS = re.compile('([^\x20-\x7d]|=)')
UNSUPPORTED_TAGS = {'r128_album_gain', 'r128_track_gain'}


def sanitize_key(key):
    """
    Remove characters from key which are invalid for a Vorbis comment field name.
    See https://www.xiph.org/vorbis/doc/v-comment.html#vectorformat
    """
    return INVALID_CHARS.sub('', key)


def is_valid_key(key):
    """
    Return true if a string is a valid Vorbis comment key.
    Valid characters for Vorbis comment field names are
    ASCII 0x20 through 0x7D, 0x3D ('=') excluded.
    """
    return key and INVALID_CHARS.search(key) is None


def flac_sort_pics_after_tags(metadata_blocks):
    """
    Reorder the metadata_blocks so that all picture blocks are located after
    the first Vorbis comment block.

    Windows fails to read FLAC tags if the picture blocks are located before
    the Vorbis comments. Reordering the blocks fixes this.
    """
    # First remember all picture blocks that are located before the tag block.
    tagindex = 0
    picblocks = []
    for block in metadata_blocks:
        if block.code == mutagen.flac.VCFLACDict.code:
            tagindex = metadata_blocks.index(block)
            break
        elif block.code == mutagen.flac.Picture.code:
            picblocks.append(block)
    else:
        return  # No tags found, nothing to sort

    # Now move those picture block after the tag block, maintaining their order.
    for pic in picblocks:
        metadata_blocks.remove(pic)
        metadata_blocks.insert(tagindex, pic)


def flac_remove_empty_seektable(file):
    """Removes an existing but empty seektable from the Flac file.

    Some software has issues with files that contain an empty seek table. Since
    no seektable is also valid, remove it.
    """
    seektable = file.seektable
    if seektable and not seektable.seekpoints:
        file.metadata_blocks = [b for b in file.metadata_blocks if b != file.seektable]
        file.seektable = None


class VCommentFile(File):

    """Generic VComment-based file."""
    _File = None

    __translate = {
        'movement': 'movementnumber',
        'movementname': 'movement',
        'musicbrainz_releasetrackid': 'musicbrainz_trackid',
        'musicbrainz_trackid': 'musicbrainz_recordingid',
        'waveformatextensible_channel_mask': '~waveformatextensible_channel_mask',
    }
    __rtranslate = {v: k for k, v in __translate.items()}

    def _load(self, filename):
        log.debug("Loading file %r", filename)
        config = get_config()
        file = self._File(encode_filename(filename))
        file.tags = file.tags or {}
        metadata = Metadata()
        for origname, values in file.tags.items():
            for value in values:
                value = value.rstrip('\0')
                name = origname
                if name in {'date', 'originaldate', 'releasedate'}:
                    # YYYY-00-00 => YYYY
                    value = sanitize_date(value)
                elif name == 'performer' or name == 'comment':
                    # transform "performer=Joe Barr (Piano)" to "performer:Piano=Joe Barr"
                    name += ':'
                    if value.endswith(')'):
                        start = len(value) - 2
                        count = 1
                        while count > 0 and start > 0:
                            if value[start] == ')':
                                count += 1
                            elif value[start] == '(':
                                count -= 1
                            start -= 1
                        if start > 0:
                            name += value[start + 2:-1]
                            value = value[:start]
                elif name.startswith('rating'):
                    try:
                        name, email = name.split(':', 1)
                    except ValueError:
                        email = ''
                    if email != sanitize_key(config.setting['rating_user_email']):
                        continue
                    name = '~rating'
                    try:
                        value = str(round((float(value) * (config.setting['rating_steps'] - 1))))
                    except ValueError:
                        log.warning('Invalid rating value in %r: %s', filename, value)
                elif name == 'fingerprint' and value.startswith('MusicMagic Fingerprint'):
                    name = 'musicip_fingerprint'
                    value = value[22:]
                elif name == 'tracktotal':
                    if 'totaltracks' in file.tags:
                        continue
                    name = 'totaltracks'
                elif name == 'disctotal':
                    if 'totaldiscs' in file.tags:
                        continue
                    name = 'totaldiscs'
                elif name == 'metadata_block_picture':
                    try:
                        image = mutagen.flac.Picture(base64.standard_b64decode(value))
                        coverartimage = TagCoverArtImage(
                            file=filename,
                            tag=name,
                            types=types_from_id3(image.type),
                            comment=image.desc,
                            support_types=True,
                            data=image.data,
                            id3_type=image.type
                        )
                    except (CoverArtImageError, TypeError, ValueError, mutagen.flac.error) as e:
                        log.error("Cannot load image from %r: %s", filename, e)
                    else:
                        metadata.images.append(coverartimage)
                    continue
                elif name in self.__translate:
                    name = self.__translate[name]
                metadata.add(name, value)
        if self._File == mutagen.flac.FLAC:
            for image in file.pictures:
                try:
                    coverartimage = TagCoverArtImage(
                        file=filename,
                        tag='FLAC/PICTURE',
                        types=types_from_id3(image.type),
                        comment=image.desc,
                        support_types=True,
                        data=image.data,
                        id3_type=image.type
                    )
                except CoverArtImageError as e:
                    log.error("Cannot load image from %r: %s", filename, e)
                else:
                    metadata.images.append(coverartimage)

        # Read the unofficial COVERART tags, for backward compatibility only
        if 'metadata_block_picture' not in file.tags:
            try:
                for data in file['COVERART']:
                    try:
                        coverartimage = TagCoverArtImage(
                            file=filename,
                            tag='COVERART',
                            data=base64.standard_b64decode(data)
                        )
                    except (CoverArtImageError, TypeError, ValueError) as e:
                        log.error("Cannot load image from %r: %s", filename, e)
                    else:
                        metadata.images.append(coverartimage)
            except KeyError:
                pass
        self._info(metadata, file)
        return metadata

    def _save(self, filename, metadata):
        """Save metadata to the file."""
        log.debug("Saving file %r", filename)
        config = get_config()
        is_flac = self._File == mutagen.flac.FLAC
        file = self._File(encode_filename(filename))
        if file.tags is None:
            file.add_tags()
        if config.setting['clear_existing_tags']:
            preserve_tags = ['waveformatextensible_channel_mask']
            if not is_flac and config.setting['preserve_images']:
                preserve_tags.append('metadata_block_picture')
                preserve_tags.append('coverart')
            preserved_values = {}
            for name in preserve_tags:
                if name in file.tags and file.tags[name]:
                    preserved_values[name] = file.tags[name]
            file.tags.clear()
            for name, value in preserved_values.items():
                file.tags[name] = value
        images_to_save = list(metadata.images.to_be_saved_to_tags())
        if is_flac and (images_to_save
                or (config.setting['clear_existing_tags']
                    and not config.setting['preserve_images'])):
            file.clear_pictures()
        tags = {}

        for name, value in metadata.items():
            if name == '~rating':
                # Save rating according to http://code.google.com/p/quodlibet/wiki/Specs_VorbisComments
                user_email = sanitize_key(config.setting['rating_user_email'])
                if user_email:
                    name = 'rating:%s' % user_email
                else:
                    name = 'rating'
                value = str(float(value) / (config.setting['rating_steps'] - 1))
            # don't save private tags
            elif name.startswith("~") or not self.supports_tag(name):
                continue
            elif name.startswith('lyrics:'):
                name = 'lyrics'
            elif name in {'date', 'originaldate', 'releasedate'}:
                # YYYY-00-00 => YYYY
                value = sanitize_date(value)
            elif name.startswith('performer:') or name.startswith('comment:'):
                # transform "performer:Piano=Joe Barr" to "performer=Joe Barr (Piano)"
                name, desc = name.split(':', 1)
                if desc:
                    value += ' (%s)' % desc
            elif name == "musicip_fingerprint":
                name = "fingerprint"
                value = "MusicMagic Fingerprint%s" % value
            elif name in self.__rtranslate:
                name = self.__rtranslate[name]
            tags.setdefault(name.upper(), []).append(value.rstrip('\0'))

        if 'totaltracks' in metadata:
            tags.setdefault('TRACKTOTAL', []).append(metadata['totaltracks'])
        if 'totaldiscs' in metadata:
            tags.setdefault('DISCTOTAL', []).append(metadata['totaldiscs'])

        for image in images_to_save:
            picture = mutagen.flac.Picture()
            picture.data = image.data
            picture.mime = image.mimetype
            picture.desc = image.comment
            picture.width = image.width
            picture.height = image.height
            picture.type = image.id3_type
            if is_flac:
                # See https://xiph.org/flac/format.html#metadata_block_picture
                expected_block_size = (8 * 4 + len(picture.data)
                    + len(picture.mime)
                    + len(picture.desc.encode('UTF-8')))
                if expected_block_size > FLAC_MAX_BLOCK_SIZE:
                    log.error("Failed saving image to %r: Image size of %d bytes exceeds maximum FLAC block size of %d bytes",
                        filename, expected_block_size, FLAC_MAX_BLOCK_SIZE)
                    continue
                file.add_picture(picture)
            else:
                tags.setdefault('METADATA_BLOCK_PICTURE', []).append(
                    base64.b64encode(picture.write()).decode('ascii'))

        file.tags.update(tags)

        self._remove_deleted_tags(metadata, file.tags)

        kwargs = {}
        if is_flac:
            flac_sort_pics_after_tags(file.metadata_blocks)
            if config.setting['fix_missing_seekpoints_flac']:
                flac_remove_empty_seektable(file)
            if config.setting['remove_id3_from_flac']:
                kwargs['deleteid3'] = True
        try:
            file.save(**kwargs)
        except TypeError:
            file.save()

    def _remove_deleted_tags(self, metadata, tags):
        """Remove the tags from the file that were deleted in the UI"""
        for tag in metadata.deleted_tags:
            real_name = self._get_tag_name(tag)
            if is_valid_key(real_name) and real_name in tags:
                if real_name in {'performer', 'comment'}:
                    parts = tag.split(':', 1)
                    if len(parts) == 2:
                        tag_type_regex = re.compile(r"\(%s\)$" % re.escape(parts[1]))
                    else:
                        tag_type_regex = re.compile(r"[^)]$")
                    existing_tags = tags.get(real_name)
                    for item in existing_tags:
                        if re.search(tag_type_regex, item):
                            existing_tags.remove(item)
                    tags[real_name] = existing_tags
                else:
                    if tag in {'totaldiscs', 'totaltracks'} and tag in tags:
                        # both tag and real_name are to be deleted in this case
                        del tags[tag]
                    del tags[real_name]

    def _get_tag_name(self, name):
        if name == '~rating':
            config = get_config()
            if config.setting['rating_user_email']:
                return 'rating:%s' % config.setting['rating_user_email']
            else:
                return 'rating'
        elif name.startswith("~"):
            return None
        elif name.startswith('lyrics:'):
            return 'lyrics'
        elif name.startswith('performer:') or name.startswith('comment:'):
            return name.split(':', 1)[0]
        elif name == 'musicip_fingerprint':
            return 'fingerprint'
        elif name == 'totaltracks':
            return 'tracktotal'
        elif name == 'totaldiscs':
            return 'disctotal'
        elif name in self.__rtranslate:
            return self.__rtranslate[name]
        else:
            return name

    @classmethod
    def supports_tag(cls, name):
        return (bool(name) and name not in UNSUPPORTED_TAGS
                and (is_valid_key(name)
                    or name.startswith('comment:')
                    or name.startswith('lyrics:')
                    or name.startswith('performer:')))


class FLACFile(VCommentFile):

    """FLAC file."""
    EXTENSIONS = [".flac"]
    NAME = "FLAC"
    _File = mutagen.flac.FLAC


class OggFLACFile(VCommentFile):

    """FLAC file."""
    EXTENSIONS = [".oggflac"]
    NAME = "Ogg FLAC"
    _File = mutagen.oggflac.OggFLAC


class OggSpeexFile(VCommentFile):

    """Ogg Speex file."""
    EXTENSIONS = [".spx"]
    NAME = "Speex"
    _File = mutagen.oggspeex.OggSpeex


class OggTheoraFile(VCommentFile):

    """Ogg Theora file."""
    EXTENSIONS = [".oggtheora"]
    NAME = "Ogg Theora"
    _File = mutagen.oggtheora.OggTheora

    def _info(self, metadata, file):
        super()._info(metadata, file)
        metadata['~video'] = '1'


class OggVorbisFile(VCommentFile):

    """Ogg Vorbis file."""
    EXTENSIONS = []
    NAME = "Ogg Vorbis"
    _File = mutagen.oggvorbis.OggVorbis


class OggOpusFile(VCommentFile):

    """Ogg Opus file."""
    EXTENSIONS = [".opus"]
    NAME = "Ogg Opus"
    _File = mutagen.oggopus.OggOpus

    @classmethod
    def supports_tag(cls, name):
        if name.startswith('r128_'):
            return True
        return VCommentFile.supports_tag(name)


def OggAudioFile(filename):
    """Generic Ogg audio file."""
    options = [OggFLACFile, OggOpusFile, OggSpeexFile, OggVorbisFile]
    return guess_format(filename, options)


OggAudioFile.EXTENSIONS = [".oga"]
OggAudioFile.NAME = "Ogg Audio"
OggAudioFile.supports_tag = VCommentFile.supports_tag


def OggVideoFile(filename):
    """Generic Ogg video file."""
    options = [OggTheoraFile]
    return guess_format(filename, options)


OggVideoFile.EXTENSIONS = [".ogv"]
OggVideoFile.NAME = "Ogg Video"
OggVideoFile.supports_tag = VCommentFile.supports_tag


def OggContainerFile(filename):
    """Generic Ogg file."""
    options = [
        OggFLACFile,
        OggOpusFile,
        OggSpeexFile,
        OggTheoraFile,
        OggVorbisFile
    ]
    return guess_format(filename, options)


OggContainerFile.EXTENSIONS = [".ogg", ".ogx"]
OggContainerFile.NAME = "Ogg"
OggContainerFile.supports_tag = VCommentFile.supports_tag



# -*- coding: utf-8 -*-

# Automatically generated - don't edit.
# Use `python setup.py build_ui` to update it.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_RenamingCompatOptionsPage(object):
    def setupUi(self, RenamingCompatOptionsPage):
        RenamingCompatOptionsPage.setObjectName("RenamingCompatOptionsPage")
        RenamingCompatOptionsPage.setEnabled(True)
        RenamingCompatOptionsPage.resize(453, 332)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(RenamingCompatOptionsPage.sizePolicy().hasHeightForWidth())
        RenamingCompatOptionsPage.setSizePolicy(sizePolicy)
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(RenamingCompatOptionsPage)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.ascii_filenames = QtWidgets.QCheckBox(RenamingCompatOptionsPage)
        self.ascii_filenames.setObjectName("ascii_filenames")
        self.verticalLayout_5.addWidget(self.ascii_filenames)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.windows_compatibility = QtWidgets.QCheckBox(RenamingCompatOptionsPage)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.windows_compatibility.sizePolicy().hasHeightForWidth())
        self.windows_compatibility.setSizePolicy(sizePolicy)
        self.windows_compatibility.setObjectName("windows_compatibility")
        self.horizontalLayout.addWidget(self.windows_compatibility)
        self.btn_windows_compatibility_change = QtWidgets.QPushButton(RenamingCompatOptionsPage)
        self.btn_windows_compatibility_change.setEnabled(False)
        self.btn_windows_compatibility_change.setObjectName("btn_windows_compatibility_change")
        self.horizontalLayout.addWidget(self.btn_windows_compatibility_change)
        self.verticalLayout_5.addLayout(self.horizontalLayout)
        self.windows_long_paths = QtWidgets.QCheckBox(RenamingCompatOptionsPage)
        self.windows_long_paths.setEnabled(False)
        self.windows_long_paths.setObjectName("windows_long_paths")
        self.verticalLayout_5.addWidget(self.windows_long_paths)
        self.replace_spaces_with_underscores = QtWidgets.QCheckBox(RenamingCompatOptionsPage)
        self.replace_spaces_with_underscores.setObjectName("replace_spaces_with_underscores")
        self.verticalLayout_5.addWidget(self.replace_spaces_with_underscores)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_replace_dir_separator = QtWidgets.QLabel(RenamingCompatOptionsPage)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_replace_dir_separator.sizePolicy().hasHeightForWidth())
        self.label_replace_dir_separator.setSizePolicy(sizePolicy)
        self.label_replace_dir_separator.setObjectName("label_replace_dir_separator")
        self.horizontalLayout_2.addWidget(self.label_replace_dir_separator)
        self.replace_dir_separator = QtWidgets.QLineEdit(RenamingCompatOptionsPage)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.replace_dir_separator.sizePolicy().hasHeightForWidth())
        self.replace_dir_separator.setSizePolicy(sizePolicy)
        self.replace_dir_separator.setMaximumSize(QtCore.QSize(20, 16777215))
        self.replace_dir_separator.setText("_")
        self.replace_dir_separator.setMaxLength(1)
        self.replace_dir_separator.setObjectName("replace_dir_separator")
        self.horizontalLayout_2.addWidget(self.replace_dir_separator)
        self.verticalLayout_5.addLayout(self.horizontalLayout_2)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_5.addItem(spacerItem)
        self.example_selection_note = QtWidgets.QLabel(RenamingCompatOptionsPage)
        self.example_selection_note.setText("")
        self.example_selection_note.setWordWrap(True)
        self.example_selection_note.setObjectName("example_selection_note")
        self.verticalLayout_5.addWidget(self.example_selection_note)

        self.retranslateUi(RenamingCompatOptionsPage)
        self.windows_compatibility.toggled['bool'].connect(self.windows_long_paths.setEnabled) # type: ignore
        self.windows_compatibility.toggled['bool'].connect(self.btn_windows_compatibility_change.setEnabled) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(RenamingCompatOptionsPage)
        RenamingCompatOptionsPage.setTabOrder(self.ascii_filenames, self.windows_compatibility)
        RenamingCompatOptionsPage.setTabOrder(self.windows_compatibility, self.btn_windows_compatibility_change)
        RenamingCompatOptionsPage.setTabOrder(self.btn_windows_compatibility_change, self.windows_long_paths)
        RenamingCompatOptionsPage.setTabOrder(self.windows_long_paths, self.replace_spaces_with_underscores)
        RenamingCompatOptionsPage.setTabOrder(self.replace_spaces_with_underscores, self.replace_dir_separator)

    def retranslateUi(self, RenamingCompatOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        self.ascii_filenames.setText(_("Replace non-ASCII characters"))
        self.windows_compatibility.setText(_("Windows compatibility"))
        self.btn_windows_compatibility_change.setText(_("Customize…"))
        self.windows_long_paths.setText(_("Allow paths longer than 259 characters"))
        self.replace_spaces_with_underscores.setText(_("Replace spaces with underscores"))
        self.label_replace_dir_separator.setText(_("Replace directory separators with:"))



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2013 Michael Wiencek
# Copyright (C) 2014-2015, 2018, 2020-2022 Laurent Monin
# Copyright (C) 2016-2017 Sambhav Kothari
# Copyright (C) 2018 Vishal Choudhary
# Copyright (C) 2018, 2022-2023 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from PyQt5 import (
    QtCore,
    QtGui,
    QtWidgets,
)

from picard.collection import (
    load_user_collections,
    user_collections,
)
from picard.i18n import sort_key


class CollectionMenu(QtWidgets.QMenu):

    def __init__(self, albums, *args):
        super().__init__(*args)
        self.ids = set(a.id for a in albums)
        self._ignore_update = False
        self.update_collections()

    def update_collections(self):
        self._ignore_update = True
        self.clear()
        self.actions = []
        for id_, collection in sorted(user_collections.items(),
                                      key=lambda k_v:
                                      (sort_key(str(k_v[1])), k_v[0])):
            action = QtWidgets.QWidgetAction(self)
            action.setDefaultWidget(CollectionMenuItem(self, collection))
            self.addAction(action)
            self.actions.append(action)
        self._ignore_update = False
        self.addSeparator()
        self.refresh_action = self.addAction(_("Refresh List"))
        self.hovered.connect(self.update_highlight)

    def refresh_list(self):
        self.refresh_action.setEnabled(False)
        load_user_collections(self.update_collections)

    def mouseReleaseEvent(self, event):
        # Not using self.refresh_action.triggered because it closes the menu
        if self.actionAt(event.pos()) == self.refresh_action and self.refresh_action.isEnabled():
            self.refresh_list()

    def update_highlight(self, action):
        if self._ignore_update:
            return
        for a in self.actions:
            a.defaultWidget().set_active(a == action)

    def update_active_action_for_widget(self, widget):
        if self._ignore_update:
            return
        for action in self.actions:
            action_widget = action.defaultWidget()
            is_active = action_widget == widget
            if is_active:
                self._ignore_hover = True
                self.setActiveAction(action)
                self._ignore_hover = False
            action_widget.set_active(is_active)


class CollectionMenuItem(QtWidgets.QWidget):

    def __init__(self, menu, collection):
        super().__init__()
        self.menu = menu
        self.active = False
        self._setup_layout(menu, collection)
        self._setup_colors()

    def _setup_layout(self, menu, collection):
        layout = QtWidgets.QVBoxLayout(self)
        style = self.style()
        layout.setContentsMargins(
            style.pixelMetric(QtWidgets.QStyle.PixelMetric.PM_LayoutLeftMargin),
            style.pixelMetric(QtWidgets.QStyle.PixelMetric.PM_FocusFrameVMargin),
            style.pixelMetric(QtWidgets.QStyle.PixelMetric.PM_LayoutRightMargin),
            style.pixelMetric(QtWidgets.QStyle.PixelMetric.PM_FocusFrameVMargin))
        self.checkbox = CollectionCheckBox(self, menu, collection)
        layout.addWidget(self.checkbox)

    def _setup_colors(self):
        palette = self.palette()
        self.text_color = palette.text().color()
        self.highlight_color = palette.highlightedText().color()

    def set_active(self, active):
        self.active = active
        palette = self.palette()
        textcolor = self.highlight_color if active else self.text_color
        palette.setColor(QtGui.QPalette.ColorRole.WindowText, textcolor)
        self.checkbox.setPalette(palette)

    def enterEvent(self, e):
        self.menu.update_active_action_for_widget(self)

    def leaveEvent(self, e):
        self.set_active(False)

    def paintEvent(self, e):
        painter = QtWidgets.QStylePainter(self)
        option = QtWidgets.QStyleOptionMenuItem()
        option.initFrom(self)
        option.state = QtWidgets.QStyle.StateFlag.State_None
        if self.isEnabled():
            option.state |= QtWidgets.QStyle.StateFlag.State_Enabled
        if self.active:
            option.state |= QtWidgets.QStyle.StateFlag.State_Selected
        painter.drawControl(QtWidgets.QStyle.ControlElement.CE_MenuItem, option)


class CollectionCheckBox(QtWidgets.QCheckBox):

    def __init__(self, parent, menu, collection):
        self.menu = menu
        self.collection = collection
        super().__init__(self.label(), parent)

        releases = collection.releases & menu.ids
        if len(releases) == len(menu.ids):
            self.setCheckState(QtCore.Qt.CheckState.Checked)
        elif not releases:
            self.setCheckState(QtCore.Qt.CheckState.Unchecked)
        else:
            self.setCheckState(QtCore.Qt.CheckState.PartiallyChecked)

    def nextCheckState(self):
        ids = self.menu.ids
        if ids & self.collection.pending:
            return
        diff = ids - self.collection.releases
        if diff:
            self.collection.add_releases(diff, self.updateText)
            self.setCheckState(QtCore.Qt.CheckState.Checked)
        else:
            self.collection.remove_releases(ids & self.collection.releases, self.updateText)
            self.setCheckState(QtCore.Qt.CheckState.Unchecked)

    def updateText(self):
        self.setText(self.label())

    def label(self):
        c = self.collection
        return ngettext("%(name)s (%(count)i release)", "%(name)s (%(count)i releases)", c.size) % {
            'name': c.name,
            'count': c.size,
        }



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2014, 2018, 2020-2021 Laurent Monin
# Copyright (C) 2017 Sambhav Kothari
# Copyright (C) 2020-2021 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from io import BytesIO
import struct

from picard.util.bitreader import LSBBitReader


class IdentificationError(Exception):
    pass


class NotEnoughData(IdentificationError):
    pass


class UnrecognizedFormat(IdentificationError):
    pass


class UnexpectedError(IdentificationError):
    pass


class IdentifyImageType:
    mime = ''
    extension = ''
    w = -1
    h = -1

    def __init__(self, data):
        self.data = data
        self.datalen = len(self.data)
        if self.datalen < 16:
            raise NotEnoughData("Not enough data")

    def read(self):
        self._read()
        return self._result()

    def _result(self):
        return (int(self.w), int(self.h), self.mime, self.extension, self.datalen)

    def match(self):
        raise NotImplementedError

    def _read(self):
        raise NotImplementedError

    @classmethod
    def all_extensions(cls):
        return [cls.extension]


class IdentifyJPEG(IdentifyImageType):
    mime = 'image/jpeg'
    extension = '.jpg'

    def match(self):
        # http://en.wikipedia.org/wiki/JPEG
        return self.data[:2] == b'\xFF\xD8'  # Start Of Image (SOI) marker

    @classmethod
    def all_extensions(cls):
        return [cls.extension, '.jpeg']

    def _read(self):
        jpeg = BytesIO(self.data)
        # skip SOI
        jpeg.read(2)
        b = jpeg.read(1)
        try:
            # https://en.wikibooks.org/wiki/JPEG_-_Idea_and_Practice/The_header_part
            # https://www.disktuna.com/list-of-jpeg-markers/
            # https://de.wikipedia.org/wiki/JPEG_File_Interchange_Format
            SOF_markers = {
                0xC0, 0xC1, 0xC2, 0xC3,
                0xC5, 0xC6, 0xC7,
                0xC9, 0xCA, 0xCB,
                0xCD, 0xCE, 0xCF
            }
            while b and ord(b) != 0xDA:  # Start Of Scan (SOS)
                while ord(b) != 0xFF:
                    b = jpeg.read(1)
                while ord(b) == 0xFF:
                    b = jpeg.read(1)
                if ord(b) in SOF_markers:
                    jpeg.read(2)  # parameter length (2 bytes)
                    jpeg.read(1)  # data precision (1 byte)
                    self.h, self.w = struct.unpack('>HH', jpeg.read(4))
                    break
                else:
                    # read 2 bytes as integer
                    length = int(struct.unpack('>H', jpeg.read(2))[0])
                    # skip data
                    jpeg.read(length - 2)
                b = jpeg.read(1)
        except struct.error:
            pass
        except ValueError:
            pass


class IdentifyGIF(IdentifyImageType):
    mime = 'image/gif'
    extension = '.gif'

    def match(self):
        # http://en.wikipedia.org/wiki/Graphics_Interchange_Format
        return self.data[:6] in {b'GIF87a', b'GIF89a'}

    def _read(self):
        self.w, self.h = struct.unpack('<HH', self.data[6:10])


class IdentifyPDF(IdentifyImageType):
    mime = 'application/pdf'
    extension = '.pdf'

    def match(self):
        # PDF
        return self.data[:4] == b'%PDF'

    def _read(self):
        self.w = self.h = 0


class IdentifyPNG(IdentifyImageType):
    mime = 'image/png'
    extension = '.png'

    def match(self):
        # http://en.wikipedia.org/wiki/Portable_Network_Graphics
        # http://www.w3.org/TR/PNG/#11IHDR
        return self.data[:8] == b'\x89PNG\x0D\x0A\x1A\x0A' and self.data[12:16] == b'IHDR'

    def _read(self):
        self.w, self.h = struct.unpack('>LL', self.data[16:24])


class IdentifyWebP(IdentifyImageType):
    mime = 'image/webp'
    extension = '.webp'

    def match(self):
        return self.data[:4] == b'RIFF' and self.data[8:12] == b'WEBP'

    def _read(self):
        data = self.data
        # See https://developers.google.com/speed/webp/docs/riff_container
        format = data[12:16]
        # Simple File Format (Lossy)
        if format == b'VP8 ':
            # See https://tools.ietf.org/html/rfc6386#section-9.1
            index = data.find(b'\x9d\x01\x2a')
            if index != -1:
                if self.datalen < index + 7:
                    raise NotEnoughData("Not enough data for WebP VP8")
                self.w, self.h = struct.unpack('<HH', data[index + 3:index + 7])
                # Width and height are encoded as 14 bit integers, ignore the first 2 bits
                self.w &= 0x3fff
                self.h &= 0x3fff
            else:
                self.w, self.h = 0, 0
        # Simple File Format (Lossless)
        elif format == b'VP8L':
            if self.datalen < 25:
                raise NotEnoughData("Not enough data for WebP VP8L")
            reader = LSBBitReader(BytesIO(data[21:25]))
            self.w = reader.bits(14) + 1
            self.h = reader.bits(14) + 1
        # Extended File Format
        elif format == b'VP8X':
            if self.datalen < 30:
                raise NotEnoughData("Not enough data for WebP VP8X")
            reader = LSBBitReader(BytesIO(data[24:30]))
            self.w = reader.bits(24) + 1
            self.h = reader.bits(24) + 1
        else:
            self.h, self.w = 0, 0


TIFF_BYTE_ORDER_LSB = b'II'
TIFF_BYTE_ORDER_MSB = b'MM'
TIFF_TAG_IMAGE_LENGTH = 257
TIFF_TAG_IMAGE_WIDTH = 256
TIFF_TYPE_SHORT = 3
TIFF_TYPE_LONG = 4


class IdentifyTiff(IdentifyImageType):
    mime = 'image/tiff'
    extension = '.tiff'

    def match(self):
        return self.data[:4] == b'II*\x00' or self.data[:4] == b'MM\x00*'

    @classmethod
    def all_extensions(cls):
        return [cls.extension, '.tif']

    def _read(self):
        # See https://www.adobe.io/content/dam/udp/en/open/standards/tiff/TIFF6.pdf
        data = self.data
        self.w, self.h = 0, 0
        byte_order = data[:2]
        if byte_order == TIFF_BYTE_ORDER_LSB:
            order = '<'
        elif byte_order == TIFF_BYTE_ORDER_MSB:
            order = '>'
        else:
            raise UnexpectedError("TIFF: unexpected byte order %r" % byte_order)
        try:
            offset, = struct.unpack(order + 'I', data[4:8])
            entry_count, = struct.unpack(order + 'H', data[offset:offset + 2])
            pos = offset + 2
            for i in range(entry_count):
                field = data[pos:pos + 12]
                tag, tiff_type = struct.unpack(order + 'HH', field[:4])
                if tag == TIFF_TAG_IMAGE_WIDTH:
                    self.w = self._read_value(tiff_type, order, field[8:12])
                    if self.h:
                        return
                elif tag == TIFF_TAG_IMAGE_LENGTH:
                    self.h = self._read_value(tiff_type, order, field[8:12])
                    if self.w:
                        return
                pos += 12
        except struct.error:
            pass

    @staticmethod
    def _read_value(tiff_type, order, data):
        if tiff_type == TIFF_TYPE_LONG:
            value = data[:4]
            struct_format = order + 'I'
        elif tiff_type == TIFF_TYPE_SHORT:
            value = data[:2]
            struct_format = order + 'H'
        else:
            raise UnexpectedError("TIFF: unexpected field type %s" % tiff_type)
        return struct.unpack(struct_format, value)[0]


knownimagetypes = (
    IdentifyJPEG,
    IdentifyPNG,
    IdentifyPDF,
    IdentifyGIF,
    IdentifyWebP,
    IdentifyTiff,
)


def identify(data):
    """Parse data for jpg, gif, png, webp, tiff and pdf metadata
    If successfully recognized, it returns a tuple with:
        - width
        - height
        - mimetype
        - extension
        - data length
    Exceptions:
        - `NotEnoughData` if data has less than 16 bytes.
        - `UnrecognizedFormat` if data isn't recognized as a known format.
        - `UnexpectedError` if unhandled cases (shouldn't happen).
        - `IdentificationError` is parent class for all preceding exceptions.
    """
    for cls in knownimagetypes:
        obj = cls(data)
        if obj.match():
            return obj.read()

    raise UnrecognizedFormat("Unrecognized image data")


def supports_mime_type(mime):
    return any(cls.mime == mime for cls in knownimagetypes)


def get_supported_extensions():
    for cls in knownimagetypes:
        yield from cls.all_extensions()



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2007-2008 Lukáš Lalinský
# Copyright (C) 2008 Will
# Copyright (C) 2009, 2019-2022 Philipp Wolfer
# Copyright (C) 2011, 2013 Michael Wiencek
# Copyright (C) 2013, 2019 Wieland Hoffmann
# Copyright (C) 2013-2014, 2018, 2020-2022 Laurent Monin
# Copyright (C) 2016 Rahul Raturi
# Copyright (C) 2016-2018 Sambhav Kothari
# Copyright (C) 2017 Antonio Larrosa
# Copyright (C) 2018, 2023 Bob Swift
# Copyright (C) 2021 Gabriel Ferreira
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


import os.path

from PyQt5 import (
    QtCore,
    QtWidgets,
)
from PyQt5.QtCore import QStandardPaths

from picard.config import (
    BoolOption,
    TextOption,
    get_config,
)
from picard.const.languages import UI_LANGUAGES
from picard.const.sys import IS_MACOS
from picard.i18n import sort_key

from picard.ui.options import (
    OptionsPage,
    register_options_page,
)
from picard.ui.theme import (
    AVAILABLE_UI_THEMES,
    OS_SUPPORTS_THEMES,
    UiTheme,
)
from picard.ui.ui_options_interface import Ui_InterfaceOptionsPage


_default_starting_dir = QStandardPaths.writableLocation(QStandardPaths.StandardLocation.HomeLocation)


class InterfaceOptionsPage(OptionsPage):

    NAME = 'interface'
    TITLE = N_("User Interface")
    PARENT = None
    SORT_ORDER = 80
    ACTIVE = True
    HELP_URL = "/config/options_interface.html"

    options = [
        BoolOption('setting', 'toolbar_show_labels', True),
        BoolOption('setting', 'toolbar_multiselect', False),
        BoolOption('setting', 'show_menu_icons', True if not IS_MACOS else False),  # On macOS it is not common that the global menu shows icons
        BoolOption('setting', 'builtin_search', True),
        BoolOption('setting', 'use_adv_search_syntax', False),
        BoolOption('setting', 'show_new_user_dialog', True),
        BoolOption('setting', 'quit_confirmation', True),
        BoolOption('setting', 'file_save_warning', True),
        TextOption('setting', 'ui_language', ''),
        TextOption('setting', 'ui_theme', str(UiTheme.DEFAULT)),
        BoolOption('setting', 'filebrowser_horizontal_autoscroll', True),
        BoolOption('setting', 'starting_directory', False),
        TextOption('setting', 'starting_directory_path', _default_starting_dir),
        TextOption('setting', 'load_image_behavior', 'append'),
    ]

    # Those are labels for theme display
    _UI_THEME_LABELS = {
        UiTheme.DEFAULT: {
            'label': N_("Default"),
            'desc': N_("The default color scheme based on the operating system display settings"),
        },
        UiTheme.DARK: {
            'label': N_("Dark"),
            'desc': N_("A dark display theme"),
        },
        UiTheme.LIGHT: {
            'label': N_("Light"),
            'desc': N_("A light display theme"),
        },
        UiTheme.SYSTEM: {
            'label': N_("System"),
            'desc': N_("The Qt5 theme configured in the desktop environment"),
        },
    }

    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_InterfaceOptionsPage()
        self.ui.setupUi(self)

        self.ui.ui_theme.clear()
        for theme in AVAILABLE_UI_THEMES:
            label = self._UI_THEME_LABELS[theme]['label']
            desc = self._UI_THEME_LABELS[theme]['desc']
            self.ui.ui_theme.addItem(_(label), theme)
            idx = self.ui.ui_theme.findData(theme)
            self.ui.ui_theme.setItemData(idx, _(desc), QtCore.Qt.ItemDataRole.ToolTipRole)
        self.ui.ui_theme.setCurrentIndex(self.ui.ui_theme.findData(UiTheme.DEFAULT))

        self.ui.ui_language.addItem(_("System default"), '')
        language_list = [(lang[0], lang[1], gettext_constants(lang[2])) for lang in UI_LANGUAGES]

        def fcmp(x):
            return sort_key(x[2])
        for lang_code, native, translation in sorted(language_list, key=fcmp):
            if native and native != translation:
                name = '%s (%s)' % (translation, native)
            else:
                name = translation
            self.ui.ui_language.addItem(name, lang_code)
        self.ui.starting_directory.toggled.connect(
            self.ui.starting_directory_path.setEnabled
        )
        self.ui.starting_directory.toggled.connect(
            self.ui.starting_directory_browse.setEnabled
        )
        self.ui.starting_directory_browse.clicked.connect(self.starting_directory_browse)

        if not OS_SUPPORTS_THEMES:
            self.ui.ui_theme_container.hide()

    def load(self):
        config = get_config()
        self.ui.toolbar_show_labels.setChecked(config.setting['toolbar_show_labels'])
        self.ui.toolbar_multiselect.setChecked(config.setting['toolbar_multiselect'])
        self.ui.show_menu_icons.setChecked(config.setting['show_menu_icons'])
        self.ui.builtin_search.setChecked(config.setting['builtin_search'])
        self.ui.use_adv_search_syntax.setChecked(config.setting['use_adv_search_syntax'])
        self.ui.new_user_dialog.setChecked(config.setting['show_new_user_dialog'])
        self.ui.quit_confirmation.setChecked(config.setting['quit_confirmation'])
        self.ui.file_save_warning.setChecked(config.setting['file_save_warning'])
        current_ui_language = config.setting['ui_language']
        self.ui.ui_language.setCurrentIndex(self.ui.ui_language.findData(current_ui_language))
        self.ui.filebrowser_horizontal_autoscroll.setChecked(config.setting['filebrowser_horizontal_autoscroll'])
        self.ui.starting_directory.setChecked(config.setting['starting_directory'])
        self.ui.starting_directory_path.setText(config.setting['starting_directory_path'])
        current_theme = UiTheme(config.setting['ui_theme'])
        self.ui.ui_theme.setCurrentIndex(self.ui.ui_theme.findData(current_theme))

    def save(self):
        config = get_config()
        config.setting['toolbar_show_labels'] = self.ui.toolbar_show_labels.isChecked()
        config.setting['toolbar_multiselect'] = self.ui.toolbar_multiselect.isChecked()
        config.setting['show_menu_icons'] = self.ui.show_menu_icons.isChecked()
        self.tagger.enable_menu_icons(config.setting['show_menu_icons'])
        config.setting['builtin_search'] = self.ui.builtin_search.isChecked()
        config.setting['use_adv_search_syntax'] = self.ui.use_adv_search_syntax.isChecked()
        config.setting['show_new_user_dialog'] = self.ui.new_user_dialog.isChecked()
        config.setting['quit_confirmation'] = self.ui.quit_confirmation.isChecked()
        config.setting['file_save_warning'] = self.ui.file_save_warning.isChecked()
        self.tagger.window.update_toolbar_style()
        new_theme_setting = str(self.ui.ui_theme.itemData(self.ui.ui_theme.currentIndex()))
        new_language = self.ui.ui_language.itemData(self.ui.ui_language.currentIndex())
        restart_warning = None
        if new_theme_setting != config.setting['ui_theme']:
            restart_warning_title = _("Theme changed")
            restart_warning = _("You have changed the application theme. You have to restart Picard in order for the change to take effect.")
            if new_theme_setting == str(UiTheme.SYSTEM):
                restart_warning += '\n\n' + _(
                    'Please note that using the system theme might cause the user interface to be not shown correctly. '
                    'If this is the case select the "Default" theme option to use Picard\'s default theme again.'
                )
        elif new_language != config.setting['ui_language']:
            restart_warning_title = _("Language changed")
            restart_warning = _("You have changed the interface language. You have to restart Picard in order for the change to take effect.")
        if restart_warning:
            dialog = QtWidgets.QMessageBox(
                QtWidgets.QMessageBox.Icon.Information,
                restart_warning_title,
                restart_warning,
                QtWidgets.QMessageBox.StandardButton.Ok,
                self)
            dialog.exec_()
        config.setting['ui_theme'] = new_theme_setting
        config.setting['ui_language'] = self.ui.ui_language.itemData(self.ui.ui_language.currentIndex())
        config.setting['filebrowser_horizontal_autoscroll'] = self.ui.filebrowser_horizontal_autoscroll.isChecked()
        config.setting['starting_directory'] = self.ui.starting_directory.isChecked()
        config.setting['starting_directory_path'] = os.path.normpath(self.ui.starting_directory_path.text())

    def starting_directory_browse(self):
        item = self.ui.starting_directory_path
        path = QtWidgets.QFileDialog.getExistingDirectory(self, "", item.text())
        if path:
            path = os.path.normpath(path)
            item.setText(path)


register_options_page(InterfaceOptionsPage)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2021, 2023 Bob Swift
# Copyright (C) 2021-2022 Laurent Monin
# Copyright (C) 2021-2023 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from copy import deepcopy
import datetime
from enum import (
    IntEnum,
    unique,
)
import os
from typing import Mapping
import uuid

import yaml

from PyQt5 import (
    QtCore,
    QtWidgets,
)

from picard import log
from picard.const import (
    DEFAULT_SCRIPT_NAME,
    SCRIPT_LANGUAGE_VERSION,
)
from picard.util import make_filename_from_title


@unique
class PicardScriptType(IntEnum):
    """Picard Script object types
    """
    BASE = 0
    TAGGER = 1
    FILENAMING = 2


class ScriptImportExportError(Exception):
    def __init__(self, *args, format=None, filename=None, error_msg=None):
        self.format = format
        self.filename = filename
        self.error_msg = error_msg


class ScriptImportError(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class MultilineLiteral(str):
    @staticmethod
    def yaml_presenter(dumper, data):
        if data:
            data = data.rstrip() + '\n'
        return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')


yaml.add_representer(MultilineLiteral, MultilineLiteral.yaml_presenter)


class PicardScript():
    """Base class for Picard script objects.
    """
    # Base class developed to support future tagging script class as possible replacement for currently used tuples in config.setting["list_of_scripts"].

    TYPE = PicardScriptType.BASE
    OUTPUT_FIELDS = ('title', 'script_language_version', 'script', 'id')

    # Don't automatically trigger changing the `script_last_updated` property when updating these properties.
    _last_updated_ignore_list = {'last_updated', 'id'}

    def __init__(self, script='', title='', id=None, last_updated=None, script_language_version=None):
        """Base class for Picard script objects

        Args:
            script (str): Text of the script.
            title (str): Title of the script.
            id (str): ID code for the script. Defaults to a system generated uuid.
            last_updated (str): The UTC date and time when the script was last updated. Defaults to current date/time.
        """
        self.title = title if title else DEFAULT_SCRIPT_NAME
        self.script = script
        if not id:
            self._set_new_id()
        else:
            self.id = id
        if last_updated is None:
            self.update_last_updated()
        else:
            self.last_updated = last_updated
        if script_language_version is None or not script_language_version:
            self.script_language_version = SCRIPT_LANGUAGE_VERSION
        else:
            self.script_language_version = script_language_version

    def _set_new_id(self):
        """Sets the ID of the script to a new system generated uuid.
        """
        self.id = str(uuid.uuid4())

    def __getitem__(self, setting):
        """A safe way of getting the value of the specified property setting, because
        it handles missing properties by returning None rather than raising an exception.

        Args:
            setting (str): The setting whose value to return.

        Returns:
            any: The value of the specified setting, or None if the setting was not found.
        """
        if hasattr(self, setting):
            value = getattr(self, setting)
            if isinstance(value, str):
                value = value.strip()
            return value
        return None

    @property
    def script(self):
        return self._script

    @script.setter
    def script(self, value):
        self._script = MultilineLiteral(value)

    @staticmethod
    def make_last_updated():
        """Provide consistently formatted last updated string.

        Returns:
            str: Last updated string from current date and time
        """
        return datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC')

    def update_last_updated(self):
        """Update the last updated attribute to the current UTC date and time.
        """
        self.last_updated = self.make_last_updated()

    def update_script_setting(self, **kwargs):
        """Updates the value of the specified properties.

        Args:
            **kwargs: Properties to update in the form `property=value`.
        """
        self.update_from_dict(kwargs)

    def update_from_dict(self, settings):
        """Updates the values of the properties based on the contents of the dictionary provided.
        Properties in the `settings` dictionary that are not found in the script object are ignored.

        Args:
            settings (dict): The settings to update.
        """
        updated = False
        for key, value in settings.items():
            if value is not None and hasattr(self, key):
                if isinstance(value, str):
                    value = value.strip()
                setattr(self, key, value)
                if key not in self._last_updated_ignore_list:
                    updated = True
        # Don't overwrite `last_updated` with a generated value if a last_updated value has been provided.
        if updated and 'last_updated' not in settings:
            self.update_last_updated()

    def to_dict(self):
        """Generate a dictionary containing the object's OUTPUT_FIELDS.

        Returns:
            dict: Dictionary of the object's OUTPUT_FIELDS
        """
        items = {key: getattr(self, key) for key in self.OUTPUT_FIELDS}
        items['description'] = str(items['description'])
        items['script'] = str(items['script'])
        return items

    def export_script(self, parent=None):
        """Export the script to a file.
        """
        # return _export_script_dialog(script_item=self, parent=parent)
        FILE_ERROR_EXPORT = N_('Error exporting file "%(filename)s": %(error)s.')

        default_script_directory = os.path.normpath(QtCore.QStandardPaths.writableLocation(QtCore.QStandardPaths.StandardLocation.DocumentsLocation))
        default_script_extension = "ptsp"
        script_filename = ".".join((self.filename, default_script_extension))
        default_path = os.path.normpath(os.path.join(default_script_directory, script_filename))

        dialog_title = _("Export Script File")
        dialog_file_types = self._get_dialog_filetypes()
        options = QtWidgets.QFileDialog.Options()
        filename, file_type = QtWidgets.QFileDialog.getSaveFileName(parent, dialog_title, default_path, dialog_file_types, options=options)
        if not filename:
            return False
        # Fix issue where Qt may set the extension twice
        (name, ext) = os.path.splitext(filename)
        if ext and str(name).endswith('.' + ext):
            filename = name
        log.debug("Exporting script file: %s", filename)
        if file_type == self._file_types()['package']:
            script_text = self.to_yaml()
        else:
            script_text = self.script + "\n"
        try:
            with open(filename, 'w', encoding='utf-8') as o_file:
                o_file.write(script_text)
        except OSError as error:
            raise ScriptImportExportError(format=FILE_ERROR_EXPORT, filename=filename, error_msg=error.strerror)
        dialog = QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.Icon.Information,
            _("Export Script"),
            _("Script successfully exported to %s") % filename,
            QtWidgets.QMessageBox.StandardButton.Ok,
            parent
        )
        dialog.exec_()
        return True

    @classmethod
    def import_script(cls, parent=None):
        """Import a script from a file.
        """
        FILE_ERROR_IMPORT = N_('Error importing "%(filename)s": %(error)s')
        FILE_ERROR_DECODE = N_('Error decoding "%(filename)s": %(error)s')

        dialog_title = _("Import Script File")
        dialog_file_types = cls._get_dialog_filetypes()
        default_script_directory = os.path.normpath(QtCore.QStandardPaths.writableLocation(QtCore.QStandardPaths.StandardLocation.DocumentsLocation))
        options = QtWidgets.QFileDialog.Options()
        filename, file_type = QtWidgets.QFileDialog.getOpenFileName(parent, dialog_title, default_script_directory, dialog_file_types, options=options)
        if not filename:
            return None
        log.debug("Importing script file: %s", filename)
        try:
            with open(filename, 'r', encoding='utf-8') as i_file:
                file_content = i_file.read()
        except OSError as error:
            raise ScriptImportExportError(format=FILE_ERROR_IMPORT, filename=filename, error_msg=error.strerror)
        if not file_content.strip():
            raise ScriptImportExportError(format=FILE_ERROR_IMPORT, filename=filename, error_msg=N_("The file was empty"))
        if file_type == cls._file_types()['package']:
            try:
                return cls().create_from_yaml(file_content)
            except ScriptImportError as error:
                raise ScriptImportExportError(format=FILE_ERROR_DECODE, filename=filename, error_msg=error)
        else:
            return cls(
                title=_("Imported from %s") % filename,
                script=file_content.strip()
            )

    @classmethod
    def create_from_dict(cls, script_dict, create_new_id=True):
        """Creates an instance based on the contents of the dictionary provided.
        Properties in the dictionary that are not found in the script object are ignored.

        Args:
            script_dict (dict): Dictionary containing the property settings.
            create_new_id (bool, optional): Determines whether a new ID is generated. Defaults to True.

        Returns:
            object: An instance of the class, populated from the property settings in the dictionary provided.
        """
        new_object = cls()
        if not isinstance(script_dict, Mapping):
            raise ScriptImportError(N_("Argument is not a dictionary"))
        if 'title' not in script_dict or 'script' not in script_dict:
            raise ScriptImportError(N_("Invalid script package"))
        new_object.update_from_dict(script_dict)
        if create_new_id or not new_object['id']:
            new_object._set_new_id()
        return new_object

    def copy(self):
        """Create a copy of the current script object with updated title and last updated attributes.
        """
        new_object = deepcopy(self)
        new_object.update_script_setting(
            title=_("%s (Copy)") % self.title,
            script_language_version=SCRIPT_LANGUAGE_VERSION,
        )
        new_object._set_new_id()
        return new_object

    def to_yaml(self):
        """Converts the properties of the script object to a YAML formatted string.  Note that only property
        names listed in `OUTPUT_FIELDS` will be included in the output.

        Returns:
            str: The properties of the script object formatted as a YAML string.
        """
        items = {key: getattr(self, key) for key in self.OUTPUT_FIELDS}
        return yaml.dump(items, sort_keys=False)

    @classmethod
    def create_from_yaml(cls, yaml_string, create_new_id=True):
        """Creates an instance based on the contents of the YAML string provided.
        Properties in the YAML string that are not found in the script object are ignored.

        Args:
            yaml_string (str): YAML string containing the property settings.
            create_new_id (bool, optional): Determines whether a new ID is generated. Defaults to True.

        Returns:
            object: An instance of the class, populated from the property settings in the YAML string.
        """
        new_object = cls()
        yaml_dict = yaml.safe_load(yaml_string)
        if not isinstance(yaml_dict, dict):
            raise ScriptImportError(N_("File content not a dictionary"))
        if 'title' not in yaml_dict or 'script' not in yaml_dict:
            raise ScriptImportError(N_("Invalid script package"))
        new_object.update_from_dict(yaml_dict)
        if create_new_id or not new_object['id']:
            new_object._set_new_id()
        return new_object

    @property
    def filename(self):
        return make_filename_from_title(self.title, _("Unnamed Script"))

    @classmethod
    def _file_types(cls):
        """Helper function to provide standard import/export file types (translated).

        Returns:
            dict: Diction of the standard file types
        """
        return {
            'all': _("All files") + " (*)",
            'script': _("Picard script files") + " (*.pts *.txt)",
            'package': _("Picard script package") + " (*.ptsp *.yaml)",
        }

    @classmethod
    def _get_dialog_filetypes(cls):
        """Helper function to build file type string used in the file dialogs.

        Returns:
            str: File type selection string
        """
        file_types = cls._file_types()
        return ";;".join((
            file_types['package'],
            file_types['script'],
            file_types['all'],
        ))


class TaggingScript(PicardScript):
    """Picard tagging script class
    """
    TYPE = PicardScriptType.TAGGER
    OUTPUT_FIELDS = ('title', 'script_language_version', 'script', 'id')

    def __init__(self, script='', title='', id=None, last_updated=None, script_language_version=None):
        """Creates a Picard tagging script object.

        Args:
            script (str): Text of the script.
            title (str): Title of the script.
            id (str): ID code for the script. Defaults to a system generated uuid.
            last_updated (str): The UTC date and time when the script was last updated. Defaults to current date/time.
        """
        super().__init__(script=script, title=title, id=id, last_updated=last_updated, script_language_version=script_language_version)


class FileNamingScript(PicardScript):
    """Picard file naming script class
    """
    TYPE = PicardScriptType.FILENAMING
    OUTPUT_FIELDS = ('title', 'description', 'author', 'license', 'version', 'last_updated', 'script_language_version', 'script', 'id')

    def __init__(
        self,
        script='',
        title='',
        id=None,
        author='',
        description='',
        license='',
        version='',
        last_updated=None,
        script_language_version=None,
        **kwargs,   # Catch additional (deprecated) arguments to avoid error in prior version config_upgrade functions.
    ):
        """Creates a Picard file naming script object.

        Args:
            script (str): Text of the script.
            title (str): Title of the script.
            id (str): ID code for the script. Defaults to a system generated uuid.
            author (str): The author of the script. Defaults to ''.
            description (str): A description of the script, typically including type of output and any required plugins or settings. Defaults to ''.
            license (str): The license under which the script is being distributed. Defaults to ''.
            version (str): Identifies the version of the script. Defaults to ''.
            last_updated (str): The UTC date and time when the script was last updated. Defaults to current date/time.
            script_language_version (str): The version of the script language supported by the script.
        """
        super().__init__(script=script, title=title, id=id, last_updated=last_updated, script_language_version=script_language_version)
        self.author = author
        self.description = description
        self.license = license
        self.version = version

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = MultilineLiteral(value)



# -*- coding: utf-8 -*-

# Automatically generated - don't edit.
# Use `python setup.py build_ui` to update it.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_TagsCompatibilityOptionsPage(object):
    def setupUi(self, TagsCompatibilityOptionsPage):
        TagsCompatibilityOptionsPage.setObjectName("TagsCompatibilityOptionsPage")
        TagsCompatibilityOptionsPage.resize(539, 705)
        self.vboxlayout = QtWidgets.QVBoxLayout(TagsCompatibilityOptionsPage)
        self.vboxlayout.setObjectName("vboxlayout")
        self.ac3_files = QtWidgets.QGroupBox(TagsCompatibilityOptionsPage)
        self.ac3_files.setObjectName("ac3_files")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.ac3_files)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.info_label = QtWidgets.QLabel(self.ac3_files)
        self.info_label.setWordWrap(True)
        self.info_label.setObjectName("info_label")
        self.verticalLayout_2.addWidget(self.info_label)
        spacerItem = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout_2.addItem(spacerItem)
        self.ac3_save_ape = QtWidgets.QRadioButton(self.ac3_files)
        self.ac3_save_ape.setChecked(True)
        self.ac3_save_ape.setObjectName("ac3_save_ape")
        self.verticalLayout_2.addWidget(self.ac3_save_ape)
        self.ac3_no_tags = QtWidgets.QRadioButton(self.ac3_files)
        self.ac3_no_tags.setObjectName("ac3_no_tags")
        self.verticalLayout_2.addWidget(self.ac3_no_tags)
        self.remove_ape_from_ac3 = QtWidgets.QCheckBox(self.ac3_files)
        self.remove_ape_from_ac3.setObjectName("remove_ape_from_ac3")
        self.verticalLayout_2.addWidget(self.remove_ape_from_ac3)
        self.vboxlayout.addWidget(self.ac3_files)
        spacerItem1 = QtWidgets.QSpacerItem(274, 41, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.vboxlayout.addItem(spacerItem1)

        self.retranslateUi(TagsCompatibilityOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(TagsCompatibilityOptionsPage)

    def retranslateUi(self, TagsCompatibilityOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        self.ac3_files.setTitle(_("AC3 files"))
        self.info_label.setText(_("Picard can save APEv2 tags to pure AC3 files, which by default do not support tagging. APEv2 tags in AC3 are supported by some players, but players not supporting AC3 files with APEv2 tags can have issues loading and playing those files. To deal with this you can choose whether to save tags to those files."))
        self.ac3_save_ape.setText(_("Save APEv2 tags"))
        self.ac3_no_tags.setText(_("Do not save tags"))
        self.remove_ape_from_ac3.setText(_("Remove APEv2 tags from AC3 files"))



# -*- coding: utf-8 -*-

# Automatically generated - don't edit.
# Use `python setup.py build_ui` to update it.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AttachedProfilesDialog(object):
    def setupUi(self, AttachedProfilesDialog):
        AttachedProfilesDialog.setObjectName("AttachedProfilesDialog")
        AttachedProfilesDialog.resize(800, 450)
        self.vboxlayout = QtWidgets.QVBoxLayout(AttachedProfilesDialog)
        self.vboxlayout.setContentsMargins(9, 9, 9, 9)
        self.vboxlayout.setSpacing(6)
        self.vboxlayout.setObjectName("vboxlayout")
        self.options_list = QtWidgets.QTableView(AttachedProfilesDialog)
        self.options_list.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.options_list.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.options_list.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.options_list.setObjectName("options_list")
        self.vboxlayout.addWidget(self.options_list)
        self.buttonBox = QtWidgets.QDialogButtonBox(AttachedProfilesDialog)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.NoButton)
        self.buttonBox.setObjectName("buttonBox")
        self.vboxlayout.addWidget(self.buttonBox)

        self.retranslateUi(AttachedProfilesDialog)
        QtCore.QMetaObject.connectSlotsByName(AttachedProfilesDialog)

    def retranslateUi(self, AttachedProfilesDialog):
        _translate = QtCore.QCoreApplication.translate
        AttachedProfilesDialog.setWindowTitle(_("Profiles Attached to Options"))



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2006 Lukáš Lalinský
# Copyright (C) 2019-2021, 2023 Philipp Wolfer
# Copyright (C) 2021 Laurent Monin
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from functools import partial

from picard.config import (
    BoolOption,
    TextOption,
    get_config,
)

from picard.ui.options import (
    OptionsPage,
    register_options_page,
)
from picard.ui.ui_options_tags_compatibility_id3 import (
    Ui_TagsCompatibilityOptionsPage,
)


class TagsCompatibilityID3OptionsPage(OptionsPage):

    NAME = 'tags_compatibility_id3'
    TITLE = N_("ID3")
    PARENT = 'tags'
    SORT_ORDER = 30
    ACTIVE = True
    HELP_URL = "/config/options_tags_compatibility_id3.html"

    options = [
        BoolOption('setting', 'write_id3v1', True),
        BoolOption('setting', 'write_id3v23', False),
        TextOption('setting', 'id3v2_encoding', 'utf-8'),
        TextOption('setting', 'id3v23_join_with', '/'),
        BoolOption('setting', 'itunes_compatible_grouping', False),
    ]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_TagsCompatibilityOptionsPage()
        self.ui.setupUi(self)
        self.ui.write_id3v23.clicked.connect(self.update_encodings)
        self.ui.write_id3v24.clicked.connect(partial(self.update_encodings, force_utf8=True))

    def load(self):
        config = get_config()
        self.ui.write_id3v1.setChecked(config.setting['write_id3v1'])
        if config.setting['write_id3v23']:
            self.ui.write_id3v23.setChecked(True)
        else:
            self.ui.write_id3v24.setChecked(True)
        if config.setting['id3v2_encoding'] == 'iso-8859-1':
            self.ui.enc_iso88591.setChecked(True)
        elif config.setting['id3v2_encoding'] == 'utf-16':
            self.ui.enc_utf16.setChecked(True)
        else:
            self.ui.enc_utf8.setChecked(True)
        self.ui.id3v23_join_with.setEditText(config.setting['id3v23_join_with'])
        self.ui.itunes_compatible_grouping.setChecked(config.setting['itunes_compatible_grouping'])
        self.update_encodings()

    def save(self):
        config = get_config()
        config.setting['write_id3v1'] = self.ui.write_id3v1.isChecked()
        config.setting['write_id3v23'] = self.ui.write_id3v23.isChecked()
        config.setting['id3v23_join_with'] = self.ui.id3v23_join_with.currentText()
        if self.ui.enc_iso88591.isChecked():
            config.setting['id3v2_encoding'] = 'iso-8859-1'
        elif self.ui.enc_utf16.isChecked():
            config.setting['id3v2_encoding'] = 'utf-16'
        else:
            config.setting['id3v2_encoding'] = 'utf-8'
        config.setting['itunes_compatible_grouping'] = self.ui.itunes_compatible_grouping.isChecked()

    def update_encodings(self, force_utf8=False):
        if self.ui.write_id3v23.isChecked():
            if self.ui.enc_utf8.isChecked():
                self.ui.enc_utf16.setChecked(True)
            self.ui.enc_utf8.setEnabled(False)
            self.ui.label_id3v23_join_with.setEnabled(True)
            self.ui.id3v23_join_with.setEnabled(True)
        else:
            self.ui.enc_utf8.setEnabled(True)
            if force_utf8:
                self.ui.enc_utf8.setChecked(True)
            self.ui.label_id3v23_join_with.setEnabled(False)
            self.ui.id3v23_join_with.setEnabled(False)


register_options_page(TagsCompatibilityID3OptionsPage)



# -*- coding: utf-8 -*-

# Automatically generated - don't edit.
# Use `python setup.py build_ui` to update it.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_EditTagDialog(object):
    def setupUi(self, EditTagDialog):
        EditTagDialog.setObjectName("EditTagDialog")
        EditTagDialog.setWindowModality(QtCore.Qt.ApplicationModal)
        EditTagDialog.resize(400, 250)
        EditTagDialog.setFocusPolicy(QtCore.Qt.StrongFocus)
        EditTagDialog.setModal(True)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(EditTagDialog)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.tag_names = QtWidgets.QComboBox(EditTagDialog)
        self.tag_names.setEditable(True)
        self.tag_names.setObjectName("tag_names")
        self.verticalLayout_2.addWidget(self.tag_names)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.value_list = QtWidgets.QListWidget(EditTagDialog)
        self.value_list.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.value_list.setTabKeyNavigation(False)
        self.value_list.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
        self.value_list.setMovement(QtWidgets.QListView.Free)
        self.value_list.setObjectName("value_list")
        self.horizontalLayout.addWidget(self.value_list)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.edit_value = QtWidgets.QPushButton(EditTagDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(100)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.edit_value.sizePolicy().hasHeightForWidth())
        self.edit_value.setSizePolicy(sizePolicy)
        self.edit_value.setMinimumSize(QtCore.QSize(100, 0))
        self.edit_value.setAutoDefault(False)
        self.edit_value.setObjectName("edit_value")
        self.verticalLayout.addWidget(self.edit_value)
        self.add_value = QtWidgets.QPushButton(EditTagDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(100)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.add_value.sizePolicy().hasHeightForWidth())
        self.add_value.setSizePolicy(sizePolicy)
        self.add_value.setMinimumSize(QtCore.QSize(100, 0))
        self.add_value.setAutoDefault(False)
        self.add_value.setObjectName("add_value")
        self.verticalLayout.addWidget(self.add_value)
        self.remove_value = QtWidgets.QPushButton(EditTagDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(120)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.remove_value.sizePolicy().hasHeightForWidth())
        self.remove_value.setSizePolicy(sizePolicy)
        self.remove_value.setMinimumSize(QtCore.QSize(120, 0))
        self.remove_value.setAutoDefault(False)
        self.remove_value.setObjectName("remove_value")
        self.verticalLayout.addWidget(self.remove_value)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Maximum)
        self.verticalLayout.addItem(spacerItem)
        self.move_value_up = QtWidgets.QPushButton(EditTagDialog)
        self.move_value_up.setText("")
        icon = QtGui.QIcon.fromTheme(":/images/16x16/go-up.png")
        self.move_value_up.setIcon(icon)
        self.move_value_up.setObjectName("move_value_up")
        self.verticalLayout.addWidget(self.move_value_up)
        self.move_value_down = QtWidgets.QPushButton(EditTagDialog)
        self.move_value_down.setText("")
        icon = QtGui.QIcon.fromTheme(":/images/16x16/go-down.png")
        self.move_value_down.setIcon(icon)
        self.move_value_down.setObjectName("move_value_down")
        self.verticalLayout.addWidget(self.move_value_down)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem1)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.buttonbox = QtWidgets.QDialogButtonBox(EditTagDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(150)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttonbox.sizePolicy().hasHeightForWidth())
        self.buttonbox.setSizePolicy(sizePolicy)
        self.buttonbox.setMinimumSize(QtCore.QSize(150, 0))
        self.buttonbox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonbox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonbox.setObjectName("buttonbox")
        self.verticalLayout_2.addWidget(self.buttonbox)

        self.retranslateUi(EditTagDialog)
        self.buttonbox.accepted.connect(EditTagDialog.accept) # type: ignore
        self.buttonbox.rejected.connect(EditTagDialog.reject) # type: ignore
        self.move_value_up.clicked.connect(EditTagDialog.move_row_up) # type: ignore
        self.move_value_down.clicked.connect(EditTagDialog.move_row_down) # type: ignore
        self.edit_value.clicked.connect(EditTagDialog.edit_value) # type: ignore
        self.add_value.clicked.connect(EditTagDialog.add_value) # type: ignore
        self.value_list.itemChanged['QListWidgetItem*'].connect(EditTagDialog.value_edited) # type: ignore
        self.remove_value.clicked.connect(EditTagDialog.remove_value) # type: ignore
        self.value_list.itemSelectionChanged.connect(EditTagDialog.value_selection_changed) # type: ignore
        self.tag_names.editTextChanged['QString'].connect(EditTagDialog.tag_changed) # type: ignore
        self.tag_names.activated['QString'].connect(EditTagDialog.tag_selected) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(EditTagDialog)
        EditTagDialog.setTabOrder(self.tag_names, self.value_list)
        EditTagDialog.setTabOrder(self.value_list, self.edit_value)
        EditTagDialog.setTabOrder(self.edit_value, self.add_value)
        EditTagDialog.setTabOrder(self.add_value, self.remove_value)
        EditTagDialog.setTabOrder(self.remove_value, self.buttonbox)

    def retranslateUi(self, EditTagDialog):
        _translate = QtCore.QCoreApplication.translate
        EditTagDialog.setWindowTitle(_("Edit Tag"))
        self.edit_value.setText(_("Edit value"))
        self.add_value.setText(_("Add value"))
        self.remove_value.setText(_("Remove value"))
        self.move_value_up.setToolTip(_("Move selected value up"))
        self.move_value_up.setAccessibleDescription(_("Move selected value up"))
        self.move_value_down.setToolTip(_("Move selected value down"))
        self.move_value_down.setAccessibleDescription(_("Move selected value down"))



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2007 Oliver Charles
# Copyright (C) 2007, 2010-2011 Lukáš Lalinský
# Copyright (C) 2007-2011, 2019-2021, 2023 Philipp Wolfer
# Copyright (C) 2011 Michael Wiencek
# Copyright (C) 2011-2012 Wieland Hoffmann
# Copyright (C) 2013-2015, 2018-2021 Laurent Monin
# Copyright (C) 2016-2017 Sambhav Kothari
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from functools import partial
import traceback

from PyQt5.QtCore import QObject

from picard import log
from picard.config import get_config
from picard.coverart.image import (
    CoverArtImageIdentificationError,
    CoverArtImageIOError,
)
from picard.coverart.providers import (
    CoverArtProvider,
    cover_art_providers,
)
from picard.metadata import register_album_metadata_processor


class CoverArt:

    def __init__(self, album, metadata, release):
        self._queue_new()
        self.album = album
        self.metadata = metadata
        self.release = release  # not used in this class, but used by providers
        self.front_image_found = False

    def __repr__(self):
        return "%s for %r" % (self.__class__.__name__, self.album)

    def retrieve(self):
        """Retrieve available cover art images for the release"""
        config = get_config()
        if config.setting['save_images_to_tags'] or config.setting['save_images_to_files']:
            self.providers = cover_art_providers()
            self.next_in_queue()
        else:
            log.debug("Cover art disabled by user options.")

    def _set_metadata(self, coverartimage, data):
        try:
            coverartimage.set_data(data)
            if coverartimage.can_be_saved_to_metadata:
                log.debug("Storing to metadata: %r [%s]",
                    coverartimage, coverartimage.imageinfo_as_string())
                self.metadata.images.append(coverartimage)
                for track in self.album._new_tracks:
                    track.metadata.images.append(coverartimage)
                # If the image already was a front image,
                # there might still be some other non-CAA front
                # images in the queue - ignore them.
                if not self.front_image_found:
                    self.front_image_found = coverartimage.is_front_image()
            else:
                log.debug("Not storing to metadata: %r [%s]",
                    coverartimage, coverartimage.imageinfo_as_string())
        except CoverArtImageIOError as e:
            self.album.error_append(e)
            self.album._finalize_loading(error=True)
            raise e
        except CoverArtImageIdentificationError as e:
            self.album.error_append(e)

    def _coverart_downloaded(self, coverartimage, data, http, error):
        """Handle finished download, save it to metadata"""
        self.album._requests -= 1

        if error:
            self.album.error_append("Coverart error: %s" % http.errorString())
        elif len(data) < 1000:
            log.warning("Not enough data, skipping %s", coverartimage)
        else:
            self._message(
                N_("Cover art of type '%(type)s' downloaded for %(albumid)s from %(host)s"),
                {
                    'type': coverartimage.types_as_string(),
                    'albumid': self.album.id,
                    'host': coverartimage.url.host(),
                },
                echo=None
            )
            try:
                self._set_metadata(coverartimage, data)
            except CoverArtImageIOError:
                # It doesn't make sense to store/download more images if we can't
                # save them in the temporary folder, abort.
                return

        self.next_in_queue()

    def next_in_queue(self):
        """Downloads next item in queue.
           If there are none left, loading of album will be finalized.
        """
        if self.album.id not in self.album.tagger.albums:
            # album removed
            return

        config = get_config()
        if (self.front_image_found
            and config.setting['save_images_to_tags']
            and not config.setting['save_images_to_files']
            and config.setting['embed_only_one_front_image']):
            # no need to continue
            self.album._finalize_loading(None)
            return

        if self._queue_empty():
            try:
                # requeue from next provider
                provider = next(self.providers)
                ret = CoverArtProvider._STARTED
                try:
                    instance = provider.cls(self)
                    if provider.enabled and instance.enabled():
                        log.debug("Trying cover art provider %s …", provider.name)
                        ret = instance.queue_images()
                    else:
                        log.debug("Skipping cover art provider %s …", provider.name)
                except BaseException:
                    log.error(traceback.format_exc())
                    raise
                finally:
                    if ret != CoverArtProvider.WAIT:
                        self.next_in_queue()
                return
            except StopIteration:
                # nothing more to do
                self.album._finalize_loading(None)
                return

        # We still have some items to try!
        coverartimage = self._queue_get()
        if not coverartimage.support_types and self.front_image_found:
            # we already have one front image, no need to try other type-less
            # sources
            log.debug("Skipping %r, one front image is already available",
                      coverartimage)
            self.next_in_queue()
            return

        # local files
        if coverartimage.url and coverartimage.url.scheme() == 'file':
            try:
                path = coverartimage.url.toLocalFile()
                with open(path, 'rb') as file:
                    self._set_metadata(coverartimage, file.read())
            except OSError as exc:
                (errnum, errmsg) = exc.args
                log.error("Failed to read %r: %s (%d)", path, errmsg, errnum)
            except CoverArtImageIOError:
                # It doesn't make sense to store/download more images if we can't
                # save them in the temporary folder, abort.
                return
            self.next_in_queue()
            return

        # on the web
        self._message(
            N_("Downloading cover art of type '%(type)s' for %(albumid)s from %(host)s …"),
            {
                'type': coverartimage.types_as_string(),
                'albumid': self.album.id,
                'host': coverartimage.url.host(),
            },
            echo=None
        )
        log.debug("Downloading %r", coverartimage)
        self.album.tagger.webservice.download_url(
            url=coverartimage.url,
            handler=partial(self._coverart_downloaded, coverartimage),
            priority=True,
        )
        self.album._requests += 1

    def queue_put(self, coverartimage):
        """Add an image to queue"""
        log.debug("Queuing cover art image %r", coverartimage)
        self.__queue.append(coverartimage)

    def _queue_get(self):
        """Get next image and remove it from queue"""
        return self.__queue.pop(0)

    def _queue_empty(self):
        """Returns True if the queue is empty"""
        return not self.__queue

    def _queue_new(self):
        """Initialize the queue"""
        self.__queue = []

    def _message(self, *args, **kwargs):
        """Display message to status bar"""
        QObject.tagger.window.set_statusbar_message(*args, **kwargs)


def _retrieve_coverart(album, metadata, release):
    """Gets all cover art URLs from the metadata and then attempts to
    download the album art. """

    coverart = CoverArt(album, metadata, release)
    log.debug("New %r", coverart)
    coverart.retrieve()


register_album_metadata_processor(_retrieve_coverart)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2013-2015, 2020-2021 Laurent Monin
# Copyright (C) 2017 Sambhav Kothari
# Copyright (C) 2018 Wieland Hoffmann
# Copyright (C) 2019-2021 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from enum import IntEnum

from picard.const import MB_ATTRIBUTES


# list of types from http://musicbrainz.org/doc/Cover_Art/Types
# order of declaration is preserved in selection box
CAA_TYPES = []
for k, v in sorted(MB_ATTRIBUTES.items(), key=lambda k_v: k_v[0]):
    if k.startswith('DB:cover_art_archive.art_type/name:'):
        CAA_TYPES.append({'name': v.lower(), 'title': v})

# pseudo type, used for the no type case
CAA_TYPES.append({'name': 'unknown', 'title': N_("Unknown")})

CAA_TYPES_TR = {}
for t in CAA_TYPES:
    CAA_TYPES_TR[t['name']] = t['title']


def translate_caa_type(name):
    if name == 'unknown':
        return _(CAA_TYPES_TR[name])
    else:
        title = CAA_TYPES_TR.get(name, name)
        return pgettext_attributes("cover_art_type", title)


# See https://id3.org/id3v2.4.0-frames, 4.14 Attached picture
class Id3ImageType(IntEnum):
    OTHER = 0
    FILE_ICON = 1
    FILE_ICON_OTHER = 2
    COVER_FRONT = 3
    COVER_BACK = 4
    LEAFLET_PAGE = 5
    MEDIA = 6
    LEAD_ARTIST = 7
    ARTIST = 8
    CONDUCTOR = 9
    BAND = 10
    COMPOSER = 11
    LYRICIST = 12
    RECORDING_DURATION = 13
    DURING_RECORDING = 14
    DURING_PERFORMANCE = 15
    VIDEO_SCREEN_CAPTURE = 16
    BRIGHT_COLOURED_FISH = 17
    ILLUSTRATION = 18
    LOGO_ARTIST = 19
    LOGO_STUDIO = 20


__ID3_IMAGE_TYPE_MAP = {
    'obi': Id3ImageType.OTHER,
    'tray': Id3ImageType.OTHER,
    'spine': Id3ImageType.OTHER,
    'sticker': Id3ImageType.OTHER,
    'other': Id3ImageType.OTHER,
    'front': Id3ImageType.COVER_FRONT,
    'back': Id3ImageType.COVER_BACK,
    'booklet': Id3ImageType.LEAFLET_PAGE,
    'track': Id3ImageType.MEDIA,
    'medium': Id3ImageType.MEDIA,
}

__ID3_REVERSE_IMAGE_TYPE_MAP = {v: k for k, v in __ID3_IMAGE_TYPE_MAP.items()}


def image_type_from_id3_num(id3type):
    return __ID3_REVERSE_IMAGE_TYPE_MAP.get(id3type, 'other')


def image_type_as_id3_num(texttype):
    return __ID3_IMAGE_TYPE_MAP.get(texttype, Id3ImageType.OTHER)


def types_from_id3(id3type):
    return [image_type_from_id3_num(id3type)]



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2006-2007 Lukáš Lalinský
# Copyright (C) 2013-2015, 2018, 2020-2021 Laurent Monin
# Copyright (C) 2014, 2019-2022 Philipp Wolfer
# Copyright (C) 2016-2017 Sambhav Kothari
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from picard.config import (
    BoolOption,
    IntOption,
    ListOption,
    TextOption,
    get_config,
)
from picard.const import QUERY_LIMIT

from picard.ui.options import (
    OptionsPage,
    register_options_page,
)
from picard.ui.ui_options_advanced import Ui_AdvancedOptionsPage


class AdvancedOptionsPage(OptionsPage):

    NAME = 'advanced'
    TITLE = N_("Advanced")
    PARENT = None
    SORT_ORDER = 90
    ACTIVE = True
    HELP_URL = "/config/options_advanced.html"

    options = [
        TextOption('setting', 'ignore_regex', ''),
        BoolOption('setting', 'ignore_hidden_files', False),
        BoolOption('setting', 'recursively_add_files', True),
        IntOption('setting', 'ignore_track_duration_difference_under', 2),
        IntOption('setting', 'query_limit', QUERY_LIMIT),
        BoolOption('setting', 'completeness_ignore_videos', False),
        BoolOption('setting', 'completeness_ignore_pregap', False),
        BoolOption('setting', 'completeness_ignore_data', False),
        BoolOption('setting', 'completeness_ignore_silence', False),
        ListOption('setting', 'compare_ignore_tags', []),
    ]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_AdvancedOptionsPage()
        self.ui.setupUi(self)
        self.init_regex_checker(self.ui.ignore_regex, self.ui.regex_error)

    def load(self):
        config = get_config()
        self.ui.ignore_regex.setText(config.setting['ignore_regex'])
        self.ui.ignore_hidden_files.setChecked(config.setting['ignore_hidden_files'])
        self.ui.recursively_add_files.setChecked(config.setting['recursively_add_files'])
        self.ui.ignore_track_duration_difference_under.setValue(config.setting['ignore_track_duration_difference_under'])
        self.ui.query_limit.setCurrentText(str(config.setting['query_limit']))
        self.ui.completeness_ignore_videos.setChecked(config.setting['completeness_ignore_videos'])
        self.ui.completeness_ignore_pregap.setChecked(config.setting['completeness_ignore_pregap'])
        self.ui.completeness_ignore_data.setChecked(config.setting['completeness_ignore_data'])
        self.ui.completeness_ignore_silence.setChecked(config.setting['completeness_ignore_silence'])
        self.ui.compare_ignore_tags.update(config.setting['compare_ignore_tags'])
        self.ui.compare_ignore_tags.set_user_sortable(False)

    def save(self):
        config = get_config()
        config.setting['ignore_regex'] = self.ui.ignore_regex.text()
        config.setting['ignore_hidden_files'] = self.ui.ignore_hidden_files.isChecked()
        config.setting['recursively_add_files'] = self.ui.recursively_add_files.isChecked()
        config.setting['ignore_track_duration_difference_under'] = self.ui.ignore_track_duration_difference_under.value()
        config.setting['query_limit'] = self.ui.query_limit.currentText()
        config.setting['completeness_ignore_videos'] = self.ui.completeness_ignore_videos.isChecked()
        config.setting['completeness_ignore_pregap'] = self.ui.completeness_ignore_pregap.isChecked()
        config.setting['completeness_ignore_data'] = self.ui.completeness_ignore_data.isChecked()
        config.setting['completeness_ignore_silence'] = self.ui.completeness_ignore_silence.isChecked()
        tags = list(self.ui.compare_ignore_tags.tags)
        if tags != config.setting['compare_ignore_tags']:
            config.setting['compare_ignore_tags'] = tags

    def restore_defaults(self):
        self.ui.compare_ignore_tags.clear()
        super().restore_defaults()


register_options_page(AdvancedOptionsPage)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2018 Sambhav Kothari
# Copyright (C) 2018, 2020-2022 Laurent Monin
# Copyright (C) 2022 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QListWidgetItem


class CheckboxListItem(QListWidgetItem):

    def __init__(self, text='', checked=False, data=None):
        super().__init__()
        self.setText(text)
        self.setFlags(self.flags() | Qt.ItemFlag.ItemIsUserCheckable)
        self.setCheckState(Qt.CheckState.Checked if checked else Qt.CheckState.Unchecked)
        self.data = data

    @property
    def checked(self):
        return self.checkState() == Qt.CheckState.Checked



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2022 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


import re

from picard.disc.utils import (
    NotSupportedTOCError,
    TocEntry,
    calculate_mb_toc_numbers,
)
from picard.util import detect_file_encoding


RE_TOC_ENTRY = re.compile(
    r"^Track (?P<num>\d+):\s+Ripped LBA (?P<start_sector>\d+) to (?P<end_sector>\d+)")


def filter_toc_entries(lines):
    """
    Take iterator of lines, return iterator of toc entries
    """
    last_track_num = 0
    for line in lines:
        m = RE_TOC_ENTRY.match(line)
        if m:
            track_num = int(m['num'])
            if last_track_num + 1 != track_num:
                raise NotSupportedTOCError(f"Non consecutive track numbers ({last_track_num} => {track_num}) in dBPoweramp log. Likely a partial rip, disc ID cannot be calculated")
            last_track_num = track_num
            yield TocEntry(track_num, int(m['start_sector']), int(m['end_sector'])-1)


def toc_from_file(path):
    """Reads dBpoweramp log files, generates MusicBrainz disc TOC listing for use as discid."""
    encoding = detect_file_encoding(path)
    with open(path, 'r', encoding=encoding) as f:
        return calculate_mb_toc_numbers(filter_toc_entries(f))



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2017 Sambhav Kothari
# Copyright (C) 2018, 2020-2021, 2023 Laurent Monin
# Copyright (C) 2018-2023 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


import re
from xml.sax.saxutils import quoteattr  # nosec: B404

from PyQt5.QtCore import QUrl

from picard import PICARD_VERSION_STR
from picard.config import get_config
from picard.const import (
    ACOUSTID_KEY,
    ACOUSTID_URL,
    MUSICBRAINZ_SERVERS,
)
from picard.util import encoded_queryargs
from picard.webservice import (
    CLIENT_STRING,
    ratecontrol,
)
from picard.webservice.utils import host_port_to_url


ratecontrol.set_minimum_delay_for_url(ACOUSTID_URL, 333)


def escape_lucene_query(text):
    return re.sub(r'([+\-&|!(){}\[\]\^"~*?:\\/])', r'\\\1', text)


def build_lucene_query(args):
    return ' '.join('%s:(%s)' % (item, escape_lucene_query(value))
                    for item, value in args.items() if value)


def _wrap_xml_metadata(data):
    return ('<?xml version="1.0" encoding="UTF-8"?>'
            '<metadata xmlns="http://musicbrainz.org/ns/mmd-2.0#">%s</metadata>'
            % data)


class APIHelper(object):
    _base_url = None

    def __init__(self, webservice, base_url=None):
        self._webservice = webservice
        if base_url is not None:
            self.base_url = base_url

    @property
    def base_url(self):
        if self._base_url is None:
            raise ValueError("base_url undefined")
        return self._base_url

    @base_url.setter
    def base_url(self, url):
        if not isinstance(url, QUrl):
            url = QUrl(url)
        self._base_url = url

    @property
    def webservice(self):
        return self._webservice

    def url_from_path(self, path):
        url = QUrl(self.base_url)
        url.setPath(url.path() + path)
        return url

    def get(self, path, handler, **kwargs):
        kwargs['url'] = self.url_from_path(path)
        kwargs['handler'] = handler
        return self._webservice.get_url(**kwargs)

    def post(self, path, data, handler, **kwargs):
        kwargs['url'] = self.url_from_path(path)
        kwargs['handler'] = handler
        kwargs['data'] = data
        kwargs['mblogin'] = kwargs.get('mblogin', True)
        return self._webservice.post_url(**kwargs)

    def put(self, path, data, handler, **kwargs):
        kwargs['url'] = self.url_from_path(path)
        kwargs['handler'] = handler
        kwargs['data'] = data
        kwargs['priority'] = kwargs.get('priority', True)
        kwargs['mblogin'] = kwargs.get('mblogin', True)
        return self._webservice.put_url(**kwargs)

    def delete(self, path, handler, **kwargs):
        kwargs['url'] = self.url_from_path(path)
        kwargs['handler'] = handler
        kwargs['priority'] = kwargs.get('priority', True)
        kwargs['mblogin'] = kwargs.get('mblogin', True)
        return self._webservice.delete_url(**kwargs)


class MBAPIHelper(APIHelper):
    @property
    def base_url(self):
        # we have to keep it dynamic since host/port can be changed via options
        config = get_config()
        host = config.setting['server_host']
        # FIXME: We should get rid of this hard coded exception and move the
        #        configuration to use proper URLs everywhere.
        port = 443 if host in MUSICBRAINZ_SERVERS else config.setting['server_port']
        self._base_url = host_port_to_url(host, port)
        self._base_url.setPath('/ws/2')
        return self._base_url

    def _get_by_id(self, entitytype, entityid, handler, inc=None, **kwargs):
        if inc:
            kwargs['unencoded_queryargs'] = kwargs.get('queryargs', {})
            kwargs['unencoded_queryargs']['inc'] = self._make_inc_arg(inc)
        return self.get(f"/{entitytype}/{entityid}", handler, **kwargs)

    def get_release_by_id(self, releaseid, handler, inc=None, **kwargs):
        return self._get_by_id('release', releaseid, handler, inc, **kwargs)

    def get_track_by_id(self, trackid, handler, inc=None, **kwargs):
        return self._get_by_id('recording', trackid, handler, inc, **kwargs)

    def lookup_discid(self, discid, handler, priority=True, important=True, refresh=False):
        inc = ('artist-credits', 'labels')
        return self._get_by_id('discid', discid, handler, inc, queryargs={'cdstubs': 'no'},
                               priority=priority, important=important, refresh=refresh)

    def _find(self, entitytype, handler, **kwargs):
        filters = {}

        limit = kwargs.pop('limit')
        if limit:
            filters['limit'] = limit

        is_search = kwargs.pop('search', False)
        if is_search:
            config = get_config()
            use_advanced_search = kwargs.pop('advanced_search', config.setting['use_adv_search_syntax'])
            if use_advanced_search:
                query = kwargs['query']
            else:
                query = escape_lucene_query(kwargs['query']).strip().lower()
                filters['dismax'] = 'true'
        else:
            query = build_lucene_query(kwargs)

        if query:
            filters['query'] = query

        return self.get(f"/{entitytype}", handler, unencoded_queryargs=filters,
                        priority=True, important=True, mblogin=False,
                        refresh=False)

    def find_releases(self, handler, **kwargs):
        return self._find('release', handler, **kwargs)

    def find_tracks(self, handler, **kwargs):
        return self._find('recording', handler, **kwargs)

    def find_artists(self, handler, **kwargs):
        return self._find('artist', handler, **kwargs)

    @staticmethod
    def _make_inc_arg(inc):
        """
        Convert an iterable to a string to be passed as inc paramater to MB

        It drops non-unique and empty elements, and sort them before joining
        them as a '+'-separated string
        """
        return '+'.join(sorted(set(str(e) for e in inc if e)))

    def _browse(self, entitytype, handler, inc=None, queryargs=None, mblogin=False):
        if queryargs is None:
            queryargs = {}
        if inc:
            queryargs['inc'] = self._make_inc_arg(inc)
        return self.get(f"/{entitytype}", handler, unencoded_queryargs=queryargs,
                        priority=True, important=True, mblogin=mblogin,
                        refresh=False)

    def browse_releases(self, handler, **kwargs):
        inc = ('media', 'labels')
        return self._browse('release', handler, inc, queryargs=kwargs)

    def browse_recordings(self, handler, inc, **kwargs):
        return self._browse('recording', handler, inc, queryargs=kwargs)

    @staticmethod
    def _xml_ratings(ratings):
        recordings = ''.join(
            '<recording id=%s><user-rating>%s</user-rating></recording>' %
            (quoteattr(i[1]), int(j)*20) for i, j in ratings.items() if i[0] == 'recording'
        )
        return _wrap_xml_metadata('<recording-list>%s</recording-list>' % recordings)

    def submit_ratings(self, ratings, handler):
        params = {'client': CLIENT_STRING}
        data = self._xml_ratings(ratings)
        return self.post("/rating", data, handler, priority=True,
                         unencoded_queryargs=params, parse_response_type='xml',
                         request_mimetype='application/xml; charset=utf-8')

    def get_collection(self, collection_id, handler, limit=100, offset=0):
        if collection_id is not None:
            inc = ('releases', 'artist-credits', 'media')
            path = f"/collection/{collection_id}/releases"
            queryargs = {
                'inc': self._make_inc_arg(inc),
                'limit': limit,
                'offset': offset,
            }
        else:
            path = '/collection'
            queryargs = None
        return self.get(path, handler, priority=True, important=True,
                        mblogin=True, unencoded_queryargs=queryargs)

    def get_collection_list(self, handler):
        return self.get_collection(None, handler)

    @staticmethod
    def _collection_request(collection_id, releases, batchsize=400):
        for i in range(0, len(releases), batchsize):
            ids = ';'.join(releases[i:i+batchsize])
            yield f"/collection/{collection_id}/releases/{ids}"

    @staticmethod
    def _get_client_queryarg():
        return {'client': CLIENT_STRING}

    def put_to_collection(self, collection_id, releases, handler):
        for path in self._collection_request(collection_id, releases):
            self.put(path, "", handler,
                     unencoded_queryargs=self._get_client_queryarg())

    def delete_from_collection(self, collection_id, releases, handler):
        for path in self._collection_request(collection_id, releases):
            self.delete(path, handler,
                        unencoded_queryargs=self._get_client_queryarg())


class AcoustIdAPIHelper(APIHelper):

    client_key = ACOUSTID_KEY
    client_version = PICARD_VERSION_STR

    def __init__(self, webservice):
        super().__init__(webservice, base_url=ACOUSTID_URL)

    def _encode_acoustid_args(self, args):
        args['client'] = self.client_key
        args['clientversion'] = self.client_version
        args['format'] = 'json'
        return '&'.join((k + '=' + v for k, v in encoded_queryargs(args).items()))

    def query_acoustid(self, handler, **args):
        body = self._encode_acoustid_args(args)
        return self.post(
            "/lookup", body, handler, priority=False, important=False,
            mblogin=False, request_mimetype='application/x-www-form-urlencoded'
        )

    @staticmethod
    def _submissions_to_args(submissions):
        config = get_config()
        args = {'user': config.setting['acoustid_apikey']}
        for i, submission in enumerate(submissions):
            for key, value in submission.args.items():
                if value:
                    args[".".join((key, str(i)))] = value
        return args

    def submit_acoustid_fingerprints(self, submissions, handler):
        args = self._submissions_to_args(submissions)
        body = self._encode_acoustid_args(args)
        return self.post(
            "/submit", body, handler, priority=True, important=False,
            mblogin=False, request_mimetype='application/x-www-form-urlencoded'
        )



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2016 Rahul Raturi
# Copyright (C) 2018, 2021-2022 Laurent Monin
# Copyright (C) 2019, 2021-2022 Philipp Wolfer
# Copyright (C) 2020 Ray Bouchard
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


"""
This can be used for basic dialogs that mostly contain a table as its core feature
"""

from abc import abstractmethod
from collections import OrderedDict

from PyQt5 import (
    QtCore,
    QtGui,
    QtWidgets,
)
from PyQt5.QtCore import pyqtSignal

from picard import log
from picard.config import get_config
from picard.i18n import sort_key
from picard.util import (
    restore_method,
    throttle,
)

from picard.ui import PicardDialog
from picard.ui.theme import theme


class ResultTable(QtWidgets.QTableWidget):

    def __init__(self, parent):
        super().__init__(parent)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
        self.setEditTriggers(QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers)
        self.horizontalHeader().setStretchLastSection(True)
        self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeMode.Stretch)
        self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeMode.Interactive)

        @throttle(1000)  # only emit scrolled signal once per second
        def emit_scrolled(x):
            parent.scrolled.emit()
        self.horizontalScrollBar().valueChanged.connect(emit_scrolled)
        self.verticalScrollBar().valueChanged.connect(emit_scrolled)
        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel)

    def prepare(self, headers):
        self.clear()
        self.setColumnCount(len(headers))
        self.setHorizontalHeaderLabels(headers)
        self.setRowCount(0)
        self.setSortingEnabled(False)


class SortableTableWidgetItem(QtWidgets.QTableWidgetItem):

    def __init__(self, sort_key):
        super().__init__()
        self.sort_key = sort_key

    def __lt__(self, other):
        return self.sort_key < other.sort_key


class TableBasedDialog(PicardDialog):

    defaultsize = QtCore.QSize(720, 360)
    scrolled = pyqtSignal()

    def __init__(self, parent):
        super().__init__(parent)
        self.setupUi()
        self.columns = None  # self.columns has to be an ordered dict, with column name as keys, and matching label as values
        self.sorting_enabled = True
        self.create_table()
        self.finished.connect(self.save_state)

    @property
    def columns(self):
        return self.__columns

    @columns.setter
    def columns(self, list_of_tuples):
        if not list_of_tuples:
            list_of_tuples = []
        self.__columns = OrderedDict(list_of_tuples)
        self.__colkeys = list(self.columns.keys())

    @property
    def table_headers(self):
        return list(self.columns.values())

    def colpos(self, colname):
        return self.__colkeys.index(colname)

    @abstractmethod
    def get_value_for_row_id(self, row, value):
        pass

    def set_table_item(self, row, colname, obj, key, sortkey=None):
        value = obj.get(key, "")
        self.set_table_item_val(row, colname, value, sortkey)

    def set_table_item_val(self, row, colname, value, sortkey=None):
        # QVariant remembers the original type of the data
        # matching comparison operator will be used when sorting
        # get() will return a string, force conversion if asked to
        if sortkey is None:
            sortkey = sort_key(value, numeric=True)
        item = SortableTableWidgetItem(sortkey)
        item.setData(QtCore.Qt.ItemDataRole.DisplayRole, value)
        pos = self.colpos(colname)
        if pos == 0:
            id = self.get_value_for_row_id(row, value)
            item.setData(QtCore.Qt.ItemDataRole.UserRole, id)
        self.table.setItem(row, pos, item)

    @abstractmethod
    def setupUi(self):
        pass

    def add_widget_to_center_layout(self, widget):
        """Update center widget with new child. If child widget exists,
        schedule it for deletion."""
        widget_item = self.center_layout.takeAt(0)
        if widget_item:
            current_widget = widget_item.widget()
            current_widget.hide()
            self.center_layout.removeWidget(current_widget)
            if current_widget != self.table:
                current_widget.deleteLater()
        self.center_layout.addWidget(widget)
        widget.show()

    def create_table_obj(self):
        return ResultTable(self)

    def create_table(self):
        self.table = self.create_table_obj()
        self.table.verticalHeader().setDefaultSectionSize(100)
        self.table.setSortingEnabled(False)
        self.table.cellDoubleClicked.connect(self.accept)
        self.table.hide()

        def enable_accept_button():
            self.accept_button.setEnabled(True)
        self.table.itemSelectionChanged.connect(enable_accept_button)

    def highlight_row(self, row):
        model = self.table.model()
        highlight_color = QtGui.QColor('LightYellow')
        if theme.is_dark_theme:
            highlight_color.setHsv(
                highlight_color.hue(),
                highlight_color.saturation(),
                int(highlight_color.lightness() * .6),
                highlight_color.alpha())
        highlight_brush = QtGui.QBrush(highlight_color)
        for column in range(0, model.columnCount()):
            index = model.index(row, column)
            model.setData(index, highlight_brush, QtCore.Qt.ItemDataRole.BackgroundRole)

    def prepare_table(self):
        self.table.prepare(self.table_headers)
        self.restore_table_header_state()

    def show_table(self, sort_column=None, sort_order=QtCore.Qt.SortOrder.DescendingOrder):
        self.add_widget_to_center_layout(self.table)
        self.table.horizontalHeader().setSortIndicatorShown(self.sorting_enabled)
        self.table.setSortingEnabled(self.sorting_enabled)
        if self.sorting_enabled and sort_column:
            self.table.sortItems(self.colpos(sort_column), sort_order)

        self.table.resizeColumnsToContents()
        self.table.resizeRowsToContents()
        self.table.setAlternatingRowColors(True)

    def accept(self):
        if self.table:
            selected_rows_user_values = []
            for idx in self.table.selectionModel().selectedRows():
                row = self.table.itemFromIndex(idx).data(QtCore.Qt.ItemDataRole.UserRole)
                selected_rows_user_values .append(row)
            self.accept_event(selected_rows_user_values)
        super().accept()

    @restore_method
    def restore_table_header_state(self):
        header = self.table.horizontalHeader()
        config = get_config()
        state = config.persist[self.dialog_header_state]
        if state:
            header.restoreState(state)
        header.setSectionResizeMode(QtWidgets.QHeaderView.ResizeMode.Interactive)
        log.debug("restore_state: %s", self.dialog_header_state)

    def save_state(self):
        if self.table:
            self.save_table_header_state()

    def save_table_header_state(self):
        state = self.table.horizontalHeader().saveState()
        config = get_config()
        config.persist[self.dialog_header_state] = state
        log.debug("save_state: %s", self.dialog_header_state)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2019-2020 Philipp Wolfer
# Copyright (C) 2020-2021 Laurent Monin
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from PyQt5 import QtWidgets

from picard.util.tags import TAG_NAMES

from picard.ui.ui_widget_taglisteditor import Ui_TagListEditor
from picard.ui.widgets.editablelistview import (
    AutocompleteItemDelegate,
    EditableListModel,
)


class TagListEditor(QtWidgets.QWidget):
    def __init__(self, parent):
        super().__init__(parent)
        self.ui = Ui_TagListEditor()
        self.ui.setupUi(self)
        list_view = self.ui.tag_list_view
        model = EditableListModel()
        model.user_sortable_changed.connect(self.on_user_sortable_changed)
        self.ui.sort_buttons.setVisible(model.user_sortable)
        list_view.setModel(model)
        list_view.setItemDelegate(AutocompleteItemDelegate(
            sorted(TAG_NAMES.keys())))

        selection = list_view.selectionModel()
        selection.selectionChanged.connect(self.on_selection_changed)
        self.on_selection_changed([], [])

    def on_selection_changed(self, selected, deselected):
        indexes = self.ui.tag_list_view.selectedIndexes()
        last_row = self.ui.tag_list_view.model().rowCount() - 1
        buttons_enabled = len(indexes) > 0
        move_up_enabled = buttons_enabled and all(i.row() != 0 for i in indexes)
        move_down_enabled = buttons_enabled and all(i.row() != last_row for i in indexes)
        self.ui.tags_remove_btn.setEnabled(buttons_enabled)
        self.ui.tags_move_up_btn.setEnabled(move_up_enabled)
        self.ui.tags_move_down_btn.setEnabled(move_down_enabled)

    def clear(self):
        self.ui.tag_list_view.update([])

    def update(self, tags):
        self.ui.tag_list_view.update(tags)

    @property
    def tags(self):
        return self.ui.tag_list_view.items

    def on_user_sortable_changed(self, user_sortable):
        self.ui.sort_buttons.setVisible(user_sortable)

    def set_user_sortable(self, user_sortable):
        self.ui.tag_list_view.model().user_sortable = user_sortable
